# !/usr/bin/python
# coding=utf-8
import re

try:
    import pymel.core as pm
except ImportError as error:
    print(__file__, error)
import pythontk as ptk

# from this package:
from mayatk.core_utils import CoreUtils, Preview


class DuplicateRadial:
    @staticmethod
    def duplicate_radial(
        objects,
        num_copies,
        start_angle=0,
        end_angle=360,
        weight_bias=0.5,
        weight_curve=4,
        rotate_axis="y",
        pivot_point=(0, 0, 0),
        translate=(0, 0, 0),
        rotate=(0, 0, 0),
        scale=(1, 1, 1),
    ):
        if rotate_axis not in ["x", "y", "z"]:
            raise ValueError("Invalid rotation axis, expected 'x', 'y', or 'z'")

        rotation_index = {"x": 0, "y": 1, "z": 2}[rotate_axis]
        total_rotation = end_angle - start_angle

        weight_factor = (
            2 * (weight_bias - 0.5) if weight_bias >= 0.5 else 2 * (0.5 - weight_bias)
        )
        originals_to_copies = {}

        for node in objects:
            # Get the parent of the original object
            parent_node = node.getParent()

            # Apply initial rotation to the object
            pm.rotate(node, rotate, r=True)

            # Scale, translate and rotate the original object
            pm.scale(node, scale, relative=True)
            pm.move(node, translate, relative=True)

            # Group the object under a new node
            group_node = pm.group(node, absolute=True)

            # Set the parent of the new group to the same as the original object
            pm.parent(group_node, parent_node)

            bb = pm.exactWorldBoundingBox(group_node)
            center_point = [(a + b) / 2 for a, b in zip(bb[:3], bb[3:])]
            original_pivot = pm.xform(group_node, q=True, rp=True)

            new_pivot_world_space = [a + b for a, b in zip(center_point, pivot_point)]
            pm.xform(group_node, piv=new_pivot_world_space)

            rotation_start = [0, 0, 0]
            rotation_start[rotation_index] = start_angle
            pm.rotate(group_node, rotation_start, r=True, os=True, fo=True)

            copies = []
            for i in range(num_copies):
                if i == 0:  # Keep the original as the first element
                    copies.append(node)
                else:
                    copy_group = pm.instance(group_node, leaf=True)[0]
                    copy = copy_group.getChildren()[0]
                    copies.append(copy)

                    x = (i - 1) / (num_copies - 1) if num_copies > 1 else 0.5
                    f_x = ptk.lerp(
                        x,
                        x**weight_curve
                        if weight_bias >= 0.5
                        else 1 - (1 - x) ** weight_curve,
                        weight_factor,
                    )

                    current_rotation = [0, 0, 0]
                    current_rotation[rotation_index] = total_rotation * f_x
                    pm.rotate(copy_group, current_rotation, r=True, os=True, fo=True)

                    t = [x * i / (num_copies - 1) for x in translate]

                    pm.move(copy_group, t)

            pm.xform(group_node, piv=original_pivot)
            originals_to_copies[node] = copies

        return originals_to_copies


class DuplicateRadialSlots:
    def __init__(self):
        self.sb = self.switchboard()
        self.ui = self.sb.duplicate_radial

        self.preview = Preview(
            self,
            self.ui.chk000,
            self.ui.b000,
            finalize_func=self.regroup_copies,
            message_func=self.sb.message_box,
        )
        self.sb.connect_multi(
            self.ui,
            "s000-16",
            "valueChanged",
            self.preview.refresh,
        )
        self.sb.connect_multi(
            self.ui,
            "chk002-4",
            "toggled",
            self.preview.refresh,
        )

    def perform_operation(self, objects):
        """Perform the radial duplication operation."""
        num_copies = self.ui.s009.value()
        start_angle = self.ui.s013.value()
        end_angle = self.ui.s014.value()
        weight_bias = self.ui.s015.value()
        weight_curve = self.ui.s016.value()
        rotate_axis = (
            "x"
            if self.ui.chk002.isChecked()
            else "y"
            if self.ui.chk003.isChecked()
            else "z"
        )
        pivot_point = (
            self.ui.s010.value(),
            self.ui.s011.value(),
            self.ui.s012.value(),
        )
        translate = (
            self.ui.s000.value(),
            self.ui.s001.value(),
            self.ui.s002.value(),
        )
        rotate = (
            self.ui.s003.value(),
            self.ui.s004.value(),
            self.ui.s005.value(),
        )
        scale = (
            self.ui.s006.value(),
            self.ui.s007.value(),
            self.ui.s008.value(),
        )

        self.copies = DuplicateRadial.duplicate_radial(
            objects,
            num_copies,
            start_angle,
            end_angle,
            weight_bias,
            weight_curve,
            rotate_axis,
            pivot_point,
            translate,
            rotate,
            scale,
        )

    def regroup_copies(self):
        """Regroup the instances under their original parent group."""
        pm.undoInfo(openChunk=True)
        for copies in self.copies.values():
            if not all(pm.objExists(copy) for copy in copies):
                # If any copy in the set doesn't exist, skip to the next set.
                continue

            first_obj_name = copies[0].name()
            name = re.sub(r"\d+$", "", first_obj_name)
            name += "_array"
            unique_name = CoreUtils.generate_unique_name(name)

            # Find the parent of the parent of the first object and use it as a parent for the new group
            original_parent = copies[0].getParent().getParent()

            for copy in copies[1:]:
                copy_group = copy.getParent()
                pm.parent(copy, world=True)
                pm.delete(copy_group)

            new_group = pm.group(copies, n=unique_name)

            # If original_parent exists then parent the new_group under original_parent
            if original_parent is not None:
                pm.parent(new_group, original_parent)

        pm.undoInfo(closeChunk=True)


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    import os
    from uitk import Switchboard

    parent = CoreUtils.get_main_window()
    ui_file = os.path.join(os.path.dirname(__file__), "duplicate_radial.ui")
    sb = Switchboard(parent, ui_location=ui_file, slot_location=DuplicateRadialSlots)

    sb.current_ui.set_attributes(WA_TranslucentBackground=True)
    sb.current_ui.set_flags(
        Tool=True, FramelessWindowHint=True, WindowStaysOnTopHint=True
    )
    sb.current_ui.set_style(theme="dark", style_class="translucentBgWithBorder")

    sb.current_ui.show(pos="screen", app_exec=True)

# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------



# !/usr/bin/python
# coding=utf-8
import unittest
import pymel.core as pm
import mayatk as mtk


class XformUtilsTest(unittest.TestCase):
    """Unit tests for the XformUtils class"""

    def setUp(self):
        """Set up test scene"""
        pm.mel.file(new=True, force=True)
        self.cube1 = pm.polyCube(name="cube1")[0]
        self.cube2 = pm.polyCube(name="cube2")[0]
        self.sph = pm.polySphere(name="sph")[0]

    def tearDown(self):
        """Clean up test scene"""
        pm.delete(self.cube1, self.cube2, self.sph)

    def test_move_to(self):
        self.assertEqual(mtk.move_to(self.cube1, self.cube2), None)

    def test_drop_to_grid(self):
        self.assertEqual(
            mtk.drop_to_grid(
                self.cube1,
                align="Min",
                origin=True,
                center_pivot=True,
                freeze_transforms=True,
            ),
            None,
        )

    def test_reset_translation(self):
        self.assertEqual(mtk.reset_translation(self.cube1), None)

    def test_set_translation_to_pivot(self):
        self.assertEqual(mtk.set_translation_to_pivot(self.cube1), None)

    def test_align_pivot_to_selection(self):
        self.assertEqual(mtk.align_pivot_to_selection(self.cube1, self.cube2), None)

    def test_aim_object_at_point(self):
        self.assertEqual(
            mtk.aim_object_at_point([self.cube1, self.cube2], (0, 15, 15)), None
        )

    def test_rotate_axis(self):
        self.assertEqual(mtk.rotate_axis([self.cube1, self.cube2], (0, 15, 15)), None)

    def test_get_orientation(self):
        self.assertEqual(
            mtk.get_orientation(self.cube1), ([1, 0, 0], [0, 1, 0], [0, 0, 1])
        )

    def test_get_distance_between_two_objects(self):
        mtk.drop_to_grid([self.cube1, self.cube2], origin=True, center_pivot=True)
        pm.move(self.cube2, 0, 0, 15)
        self.assertEqual(mtk.get_dist_between_two_objects(self.cube1, self.cube2), 15)

    def test_get_center_point(self):
        self.assertEqual(mtk.get_center_point(self.sph), (0, 0, 0))

    def test_get_bounding_box(self):
        self.assertEqual(
            mtk.get_bounding_box(self.sph, "size"),
            (2.000000238418579, 2.0, 2.0000005960464478),
        )

    def test_sort_by_bounding_box_value(self):
        self.assertEqual(
            str(mtk.sort_by_bounding_box_value(["sph.vtx[0]", "sph.f[0]"])),
            "[MeshVertex('sphShape.vtx[0]'), MeshFace('sphShape.f[0]')]",
        )

    def test_match_scale(self):
        self.assertEqual(
            mtk.match_scale(self.cube1, self.cube2, scale=False), [1.0, 1.0, 1.0]
        )

    def test_align_using_three_points(self):
        pass
        # self.assertEqual(mtk.align_using_three_points(), None)

    def test_is_overlapping(self):
        pass
        # self.assertEqual(mtk.is_overlapping(), None)

    def test_align_vertices(self):
        pass
        # self.assertEqual(mtk.align_vertices(), None)


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    mtk.clear_scroll_field_reporter()

    # Create a Test Suite
    suite = unittest.TestSuite()

    # Add the test case class to the suite
    suite.addTest(unittest.makeSuite(XformUtilsTest))

    # Run the suite
    runner = unittest.TextTestRunner(verbosity=2)
    runner.run(suite)


# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------



# !/usr/bin/python
# coding=utf-8
import os

try:
    import pymel.core as pm
except ImportError as error:
    print(__file__, error)
import pythontk as ptk

# from this package:
from mayatk.core_utils import CoreUtils


class ReferenceManager:
    def __init__(self):
        self.references = {}
        self._recursive_search = True
        self._filter_text = ""

    @property
    def workspace_root(self):
        return pm.workspace(q=True, rd=True)

    @workspace_root.setter
    def workspace_root(self, value):
        pm.workspace(dir=value)

    @property
    def recursive_search(self):
        return self._recursive_search

    @recursive_search.setter
    def recursive_search(self, value):
        self._recursive_search = value

    @property
    def filter_text(self):
        return self._filter_text

    @filter_text.setter
    def filter_text(self, value):
        self._filter_text = value

    def _get_workspace_files(self, search_root, extensions=["*.ma", "*.mb"]):
        workspace_files = ptk.get_dir_contents(
            search_root,
            returned_type="filepath",
            inc_files=extensions,
            num_threads=4,
            recursive=self._recursive_search,
        )
        if self._filter_text:
            workspace_files = ptk.filter_list(
                workspace_files, inc=[f"*{self._filter_text}*"]
            )

        return workspace_files

    def add_reference(self, namespace, file_path):
        # Validate and create reference
        if not os.path.exists(file_path):
            raise FileNotFoundError(f"File {file_path} does not exist.")
        ref_node = pm.createReference(file_path, namespace=namespace)
        self.references[
            namespace
        ] = ref_node  # Store the reference node, not just the path

    def remove_reference(self, namespace):
        if namespace in self.references:
            self.references[namespace].remove()  # Remove the reference in Maya
            del self.references[namespace]  # Remove from internal dictionary

    def resolve_file_path(self, selected_file, search_root):
        workspace_files = self._get_workspace_files(search_root)
        return next((fp for fp in workspace_files if selected_file in fp), None)


class ReferenceManagerSlots(ReferenceManager):
    def __init__(self):
        super().__init__()
        self.sb = self.switchboard()
        self.ui = self.sb.reference_manager

        self.ui.txt000.setText(self.workspace_root)
        self.ui.txt000.textChanged.connect(self.set_custom_root)
        self.ui.txt001.textChanged.connect(self.update_filtered_list)

        self.ui.list000.doubleClicked.connect(self.add_reference_from_list)
        self.ui.list000.itemClicked.connect(self.handle_item_click)
        self.ui.list000.setSelectionMode(self.sb.QAbstractItemView.MultiSelection)

        self.update_filtered_list()

    def txt000_init(self, widget):
        """ """
        widget.menu.mode = "context"
        widget.menu.add(
            "QPushButton",
            setText="Browse",
            setObjectName="b000",
            setToolTip="Open a file browser to select a root directory.",
        )
        widget.menu.add(
            "QPushButton",
            setText="Set To Workspace",
            setObjectName="b001",
            setToolTip="Set the root folder to that of the current workspace.",
        )
        widget.menu.add(
            "QCheckBox",
            setText="Recursive Search",
            setObjectName="chk000",
            setChecked=True,
            setToolTip="Also search sub-folders.",
        )

        # Connect buttons and checkbox to their respective methods
        self.ui.b000.clicked.connect(self.browse_for_root_directory)
        self.ui.b001.clicked.connect(self.set_root_to_workspace)
        self.ui.chk000.toggled.connect(self.toggle_recursive_search)

    def toggle_recursive_search(self, checked):
        self.recursive_search = checked
        self.update_filtered_list()

    def set_custom_root(self):
        new_dir = self.ui.txt000.text()
        if os.path.isdir(new_dir):
            self.custom_root = new_dir
            self.ui.txt000.setToolTip(new_dir)
            self.update_filtered_list()

    def update_filtered_list(self):
        self.filter_text = self.ui.txt001.text().strip()
        search_root = getattr(self, "custom_root", self.workspace_root)
        file_list = [
            os.path.basename(fp) for fp in self._get_workspace_files(search_root)
        ]
        self._populate_list(file_list)

    def browse_for_root_directory(self):
        selected_directory = self.sb.dir_dialog("Select a root directory")
        if selected_directory:
            self.ui.txt000.setText(selected_directory)

    def set_root_to_workspace(self):
        self.ui.txt000.setText(self.workspace_root)

    def _populate_list(self, file_list):
        self.ui.list000.clear()
        self.ui.list000.addItems(file_list)

    def handle_item_click(self, item):
        selected_file = item.text()
        search_root = getattr(self, "custom_root", self.workspace_root)
        file_path = self.resolve_file_path(selected_file, search_root)
        if file_path:
            namespace = os.path.splitext(selected_file)[0]
            if item.isSelected():
                try:
                    self.add_reference(namespace, file_path)
                except ValueError:
                    print(f"Namespace {namespace} already in use.")
            else:
                try:
                    self.remove_reference(
                        namespace
                    )  # This now calls the updated remove_reference
                except KeyError:
                    print(f"No reference found for namespace {namespace}.")

    def add_reference_from_list(self, index):
        selected_file = self.ui.list000.itemFromIndex(index).text()
        search_root = getattr(self, "custom_root", self.workspace_root)
        file_path = self.resolve_file_path(selected_file, search_root)
        if file_path:
            namespace = os.path.splitext(selected_file)[0]
            self.add_reference(namespace, file_path)


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    from uitk import Switchboard

    parent = CoreUtils.get_main_window()
    ui_file = os.path.join(os.path.dirname(__file__), "reference_manager.ui")
    sb = Switchboard(parent, ui_location=ui_file, slot_location=ReferenceManagerSlots)

    sb.current_ui.set_attributes(WA_TranslucentBackground=True)
    sb.current_ui.set_flags(FramelessWindowHint=True, WindowStaysOnTopHint=True)
    sb.current_ui.set_style(theme="dark", style_class="translucentBgWithBorder")
    sb.current_ui.header.configureButtons(minimize_button=True, hide_button=True)
    sb.current_ui.show(pos="screen", app_exec=True)
# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------



# !/usr/bin/python
# coding=utf-8
try:
    import pymel.core as pm
except ImportError as error:
    print(__file__, error)
import pythontk as ptk

# from this package:
from mayatk import core_utils


class NodeUtils:
    """ """

    @staticmethod
    def node_exists(n, search="name"):
        """Check if the node exists in the current scene.

        Parameters:
            search (str): The search parameters. valid: 'name', 'type', 'exactType'

        Returns:
            (bool)
        """
        if search == "name":
            return bool(pm.ls(n))
        elif search == "type":
            return bool(pm.ls(type=n))
        elif search == "exactType":
            return bool(pm.ls(exactType=n))

    @classmethod
    def get_type(cls, objects):
        """Get the object type as a string.

        Parameters:
            objects (str/obj/list): The object(s) to query.

        Returns:
            (str/list) The node type. A list is always returned when 'objects' is given as a list.
        """
        types = []
        for obj in pm.ls(objects):
            if cls.is_group(obj):
                typ = "group"
            elif cls.is_locator(obj):
                typ = "locator"
            else:
                typ = core_utils.Components.get_component_type(obj)
            if not typ:
                typ = pm.objectType(obj)
            types.append(typ)

        return ptk.format_return(types, objects)

    @classmethod
    def is_locator(cls, obj):
        """Check if the object is a locator.
        A locator is a transform node that has a shape node child.
        The shape node defines the appearance and behavior of the locator.

        Parameters:
            obj () = The object to query.

        Returns:
            (bool)
        """
        shape = cls.get_shape_node(obj)
        return pm.nodeType(shape) == "locator"

    @staticmethod
    def is_group(objects):
        """Determine if each of the given object(s) is a group.
        A group is defined as a transform node with children and no shape nodes directly beneath it.

        Parameters:
            objects (str/obj/list): The object(s) to query.

        Returns:
            (bool)(list) A list is always returned when 'objects' is given as a list.
        """
        result = []
        for n in pm.ls(objects):
            try:
                # Check if the object is a transform node
                is_transform = type(n) == pm.nodetypes.Transform
                # Check if the transform node does not have any shape nodes directly beneath it
                no_shapes = len(n.getShapes(noIntermediate=True)) == 0
                q = is_transform and no_shapes
            except AttributeError:
                q = False
            result.append(q)

        return ptk.format_return(result, objects)

    @classmethod
    def get_groups(cls, empty=False):
        """Get all groups in the scene.

        Parameters:
            empty (bool): Return only empty groups.

        Returns:
            (bool)
        """
        transforms = pm.ls(type="transform")

        groups = []
        for t in transforms:
            if cls.is_group(t):
                if empty:
                    children = pm.listRelatives(t, children=True)
                    if children:
                        continue
                groups.append(t)

        return groups

    @staticmethod
    def get_parent(node, all=False):
        """List the parents of an object."""
        if all:
            objects = pm.ls(node, l=1)
            return objects[0].split("|")

        try:
            return pm.listRelatives(node, parent=True, type="transform")[0]
        except IndexError:
            return None

    @staticmethod
    def get_children(node):
        """List the children of an object."""
        try:
            return pm.listRelatives(node, children=True, type="transform")
        except IndexError:
            return []

    @staticmethod
    def get_unique_children(objects):
        """Retrieves a unique list of objects' children (if any) in the scene, excluding the groups themselves.

        This function takes a list of objects in the scene and, if any object is a group, retrieves
        its children. The resulting list includes the unique children of the groups, but not the groups themselves.
        If an object is not a group, it will be included in the list.

        Parameters:
            objects (str/obj/list): A string, PyNode, or list of PyNodes representing the objects in the scene.

        Returns:
            list: A list containing the unique children of the groups (if any) and other objects.

        Example:
            >>> get_unique_children(<group>) # Returns: [nt.Transform(u'pCube1'), nt.Transform(u'pCube2')]
        """
        objects = pm.ls(objects, flatten=True)

        final_set = (
            {  # Combine the selected objects and their children into a single set.
                child
                for obj in objects
                for child in (
                    pm.listRelatives(obj, children=True, type="transform")
                    if obj.nodeType() == "transform"
                    and pm.listRelatives(obj, children=True, type="transform")
                    else (
                        [obj]
                        if obj.nodeType() != "transform"
                        or obj.listRelatives(children=True)
                        else []
                    )
                )
            }
        )
        return list(final_set)  # Convert the set back to a list.

    @staticmethod
    def get_transform_node(
        nodes, returned_type="obj", attributes=False, inc=[], exc=[]
    ):
        """Get transform node(s) or node attributes.

        Parameters:
            nodes (str/obj/list): A relative of a transform Node.
            returned_type (str): The desired returned object type. Not valid with the `attributes` parameter.
                    (valid: 'str'(default), 'obj').
            attributes (bool): Return the attributes of the node, rather then the node itself.

        Returns:
            (obj/list) node(s) or node attributes. A list is always returned when 'nodes' is given as a list.
        """
        result = []
        for node in pm.ls(nodes):
            transforms = pm.ls(node, type="transform")
            if not transforms:  # from shape
                shapeNodes = pm.ls(node, objectsOnly=1)
                transforms = pm.listRelatives(shapeNodes, parent=1)
                if not transforms:  # from history
                    try:
                        transforms = pm.listRelatives(
                            pm.listHistory(node, future=1), parent=1
                        )
                    except Exception:
                        transforms = []
            for n in transforms:
                result.append(n)

        if attributes:
            result = pm.listAttr(result, read=1, hasData=1)

        # convert element type.
        result = core_utils.CoreUtils.convert_array_type(
            result, returned_type=returned_type, flatten=True
        )
        # filter
        result = ptk.filter_list(result, inc, exc)
        # return as list if `nodes` was given as a list.
        return ptk.format_return(list(set(result)), nodes)

    @classmethod
    def get_shape_node(
        cls, nodes, returned_type="obj", attributes=False, inc=[], exc=[]
    ):
        """Get shape node(s) or node attributes.

        Parameters:
            nodes (str/obj/list): A relative of a shape Node.
            returned_type (str): The desired returned object type.
                    (valid: 'str'(default), 'obj'(shape node), 'transform'(as string), 'int'(valid only at sub-object level).
            attributes (bool): Return the attributes of the node, rather then the node itself.

        Returns:
            (obj/list) node(s) or node attributes. A list is always returned when 'nodes' is given as a list.
        """
        result = []
        for node in pm.ls(nodes):
            shapes = pm.listRelatives(
                node, children=1, shapes=1
            )  # get shape node from transform: returns list ie. [nt.Mesh('pConeShape1')]
            if not shapes:
                shapes = pm.ls(node, type="shape")
                if not shapes:  # get shape from transform
                    try:
                        transforms = pm.listRelatives(
                            pm.listHistory(node, future=1), parent=1
                        )
                        shapes = cls.get_shape_node(transforms)
                    except Exception:
                        shapes = []
            for n in shapes:
                result.append(n)

        if attributes:
            result = pm.listAttr(result, read=1, hasData=1)

        # convert element type.
        result = core_utils.CoreUtils.convert_array_type(
            result, returned_type=returned_type, flatten=True
        )
        # filter
        result = ptk.filter_list(result, inc, exc)
        # return as list if `nodes` was given as a list.
        return ptk.format_return(list(set(result)), nodes)

    @staticmethod
    def get_history_node(nodes, returned_type="obj", attributes=False, inc=[], exc=[]):
        """Get history node(s) or node attributes.

        Parameters:
            nodes (str/obj/list): A relative of a history Node.
            returned_type (str): The desired returned object type.
                    (valid: 'str'(default), 'obj'(shape node), 'transform'(as string), 'int'(valid only at sub-object level).
            attributes (bool): Return the attributes of the node, rather then the node itself.

        Returns:
            (obj/list) node(s) or node attributes. A list is always returned when 'nodes' is given as a list.
        """
        result = []
        for node in pm.ls(nodes):
            shapes = pm.listRelatives(
                node, children=1, shapes=1
            )  # get shape node from transform: returns list ie. [nt.Mesh('pConeShape1')]
            try:
                history = pm.listConnections(shapes, source=1, destination=0)[
                    -1
                ]  # get incoming connections: returns list ie. [nt.PolyCone('polyCone1')]
            except IndexError:
                try:
                    history = node.history()[-1]
                except AttributeError as error:
                    print(
                        "{} in get_history_node\n\t# Error: {} #".format(
                            __file__, error
                        )
                    )
                    continue
            result.append(history)

        if attributes:
            result = pm.listAttr(result, read=1, hasData=1)

        # convert element type.
        result = core_utils.CoreUtils.convert_array_type(
            result, returned_type=returned_type, flatten=True
        )
        # filter
        result = ptk.filter_list(result, inc, exc)
        # return as list if `nodes` was given as a list.
        return ptk.format_return(list(set(result)), nodes)

    @classmethod
    def create_render_node(
        cls,
        node_type,
        flag="asShader",
        secondary_flag="surfaceShader",
        name="",
        tex="",
        texture_node=False,
        post_command="",
        **kwargs,
    ):
        """Procedure to create the node classified as specified by the inputs.

        Parameters:
            node_type (str): The type of node to be created. ie. 'StingrayPBS' or 'aiStandardSurface'
            flag (str): A flag specifying which how to classify the node created.
                    valid:  as2DTexture, as3DTexture, asEnvTexture, asShader, asLight, asUtility
            secondary_flag (str): A secondary flag used to make decisions in combination with 'asType'
                    valid:  -asBump : defines a created texture as a bump
                                    -asNoShadingGroup : for materials; create without a shading group
                                    -asDisplacement : for anything; map the created node to a displacement material.
                                    -asUtility : for anything; do whatever the $as flag says, but also classify as a utility
                                    -asPostProcess : for any postprocess node
            name (str): The desired node name.
            tex (str): The path to a texture file for those nodes that support one.
            texture_node (bool): If not needed, the `place2dTexture` node will be deleted after creation.
            post_command (str): A command entered by the user when invoking create_render_node.
                            The command will substitute the string %node with the name of the
                            node it creates.  createRenderWindow will be closed if a command
                            is not the null string ("").
            kwargs () = Set additional node attributes after creation. ie. colorSpace='Raw', alphaIsLuminance=1, ignoreColorSpaceFileRules=1

        Returns:
            (obj) node

        Example:
            create_render_node('StingrayPBS')
            create_render_node('file', flag='as2DTexture', tex=f, texture_node=True, colorSpace='Raw', alphaIsLuminance=1, ignoreColorSpaceFileRules=1)
            create_render_node('aiSkyDomeLight', tex=pathToHdrMap, name='env', camera=0, skyRadius=0) #turn off skydome and viewport visibility.
        """
        node = pm.PyNode(
            pm.mel.createRenderNodeCB(
                "-" + flag, secondary_flag, node_type, post_command
            )
        )  # node = pm.shadingNode(typ, asTexture=True)

        if not texture_node:
            pm.delete(
                pm.listConnections(
                    node, type="place2dTexture", source=True, exactType=True
                )
            )

        if tex:
            try:
                node.fileTextureName.set(tex)
            except Exception as error:
                print("# Error:", __file__, error, "#")

        if name:
            try:
                pm.rename(node, name)

            except RuntimeError as error:
                print("# Error:", __file__, error, "#")

        cls.set_node_attributes(node, **kwargs)
        return node

    @staticmethod
    def get_connected_nodes(
        node, node_type=None, direction=None, exact=True, first_match=False
    ):
        """Finds connected nodes of a given type and direction (incoming/outgoing).

        Parameters:
            node (PyNode): The node to start searching from.
            node_type (str, optional): The node type to look for. If None, returns all connected nodes.
            direction (str, optional): 'incoming' for incoming, 'outgoing' for outgoing, None for both.
            exact (bool): Only consider nodes of the exact type. Otherwise, derived types are also considered.
            first_match (bool): Return only the first found node that matches the criteria, if any.

        Returns:
            list or obj or None: List of connected nodes or single node based on the conditions, or None if not found.
        """
        visited = set()
        stack = [node]
        filtered_nodes = []

        source, dest = {
            "incoming": (True, False),
            "outgoing": (False, True),
        }.get(direction, (True, True))

        while stack:
            current_node = stack.pop()
            visited.add(current_node)

            connected_nodes = pm.listConnections(
                current_node, s=source, d=dest, exactType=exact
            )

            for n in connected_nodes:
                if n in visited:
                    continue

                if node_type is None or pm.nodeType(n) == node_type:
                    filtered_nodes.append(n)
                    if first_match:
                        return n

                if direction is None:
                    stack.append(n)

        return filtered_nodes if not first_match else None

    @staticmethod
    def get_node_attributes(node, inc=[], exc=[], mapping=False, **kwargs):
        """Retrieves specified node's attributes along with their corresponding values, represented as a dictionary.

        Parameters:
            node (obj): The target node from which to extract attributes.
            inc (str/list): Specifies which attributes to include in the output. Any other attributes will be ignored. If there is any overlap, the exclude parameter takes priority over this.
            exc (str/list): Determines which attributes to leave out from the result.
            mapping (bool): If set to True, returns a dictionary mapping attributes to their respective values.
            kwargs: Supports additional keyword arguments that are passed to the listAttr command in Maya.

        Returns:
            dict: A dictionary that pairs each attribute (as a string) to its current value. If 'mapping' is False, only the values of the attributes are returned.
        """
        kwargs.setdefault("read", True)
        kwargs.setdefault("hasData", True)
        kwargs.setdefault("settable", True)

        attr_names = ptk.filter_list(
            pm.listAttr(node, **kwargs),
            inc,
            exc,
        )

        result = {}
        for attr_name in attr_names:
            try:
                result[attr_name] = pm.getAttr(f"{node}.{attr_name}", silent=True)
            except pm.MayaAttributeError as e:
                print(
                    f"Error encountered while extracting attribute {attr_name} from node {node}: {e}"
                )
                continue

        return result if mapping else result.values()

    @classmethod
    def set_node_attributes(cls, node, **attributes):
        """Set node attribute values. If the attribute doesn't exist, it will be created.

        Parameters:
            node (str/obj): The node to set attributes of.
            attributes (dict): Attributes and their corresponding value to set. ie. attribute_name=value
        """
        for attr, value in attributes.items():
            try:
                pm.setAttr(f"{node}.{attr}", value)
            except AttributeError:
                cls.set_node_custom_attributes(node, **{attr: value})

    @classmethod
    def get_maya_attribute_type(cls, value):
        """Gets the corresponding Maya attribute type for a given value.

        This method determines the Maya attribute type based on the type and structure of the input value.
        It supports basic data types like bool, int, float, and str, as well as more complex types like
        lists, tuples, sets, and matrices.

        Parameters:
            value: The value to determine the Maya attribute type for.

        Returns:
            str: The corresponding Maya attribute type as a string.

        Raises:
            TypeError: If the input value's type is not supported.

        Example:
            >>> get_maya_attribute_type(True) #Returns: 'bool'
            >>> get_maya_attribute_type(42) #Returns: 'long'
            >>> get_maya_attribute_type([1.0, 2.0, 3.0]) #Returns: 'double3'
            >>> get_maya_attribute_type([["a", "b"], ["c", "d"]]) #Returns: 'stringArray'
            >>> get_maya_attribute_type(Matrix([[1.0, 0.0], [0.0, 1.0]], type='float')) #Returns: 'fltMatrix'

        Notes:
            - To support additional data types, add more cases in the method as necessary.
            - Replace 'Matrix' with the correct class for a matrix in your environment.
        """
        if isinstance(value, bool):
            return "bool"
        elif isinstance(value, int):
            return "long"
        elif isinstance(value, float):
            return "double"
        elif isinstance(value, str):
            return "string"
        elif isinstance(value, (list, tuple, set)):
            element_type = cls.get_maya_attribute_type(value[0])
            length = len(value)
            # Handle compound and array cases
            if element_type in ["double", "float", "long", "short"]:
                if length == 2:
                    return f"{element_type}2"
                elif length == 3:
                    return f"{element_type}3"
                else:
                    return f"{element_type}Array"
            elif element_type == "string":
                return "stringArray"
            elif element_type == "vector":
                return "vectorArray"
            elif element_type == "point":
                return "pointArray"
            else:
                return "compound"
        # Add more cases for other data types if necessary
        elif isinstance(
            value, pm.Matrix
        ):  # Replace Matrix with the correct class for a matrix in your environment
            if value.type == "float":
                return "fltMatrix"
            elif value.type == "double":
                return "matrix"

    @classmethod
    def set_node_custom_attributes(cls, node, **attributes):
        """Set node attribute values. If the attribute doesn't exist, it will be created.

        Parameters:
            node (str/obj): The node to set attributes of.
            attributes (dict): Attributes and their corresponding value to set. ie. attribute_name=value
        """
        if isinstance(node, str):
            node = pm.PyNode(node)

        for attr, value in attributes.items():
            attr_type = cls.get_maya_attribute_type(value)

            if not pm.attributeQuery(attr, node=node, exists=True):
                if attr_type.endswith("3") or attr_type.endswith(
                    "2"
                ):  # Check if the attribute type is a compound attribute
                    node.addAttr(
                        attr, numberOfChildren=len(value), attributeType="compound"
                    )
                    component_suffixes = (
                        ["X", "Y", "Z"] if attr_type.endswith("3") else ["X", "Y"]
                    )
                    child_attr_type = attr_type[:-1]  # remove the 2 or 3 suffix.
                    for i, component in enumerate(value):
                        component_name = f"{attr}{component_suffixes[i]}"
                        node.addAttr(
                            component_name, attributeType=child_attr_type, parent=attr
                        )
                    for i, component in enumerate(
                        value
                    ):  # Separate loop to set attribute values
                        component_name = f"{attr}{component_suffixes[i]}"
                        pm.setAttr(f"{node}.{component_name}", component)
                else:
                    node.addAttr(
                        attr, defaultValue=value, keyable=True, dataType=attr_type
                    )
                    pm.setAttr(
                        f"{node}.{attr}", value
                    )  # Set attribute value immediately after creation
            else:
                if isinstance(value, (list, tuple)):  # Handle compound attributes
                    component_suffixes = (
                        ["X", "Y", "Z"] if len(value) == 3 else ["X", "Y"]
                    )
                    for i, component in enumerate(value):
                        component_name = f"{attr}{component_suffixes[i]}"
                        pm.setAttr(f"{node}.{component_name}", component)
                else:
                    pm.setAttr(f"{node}.{attr}", value)

    @staticmethod
    def connect_attributes(attr, place, file):
        """Connects a given attribute between two nodes using a specified place and file node.

        This convenience function is designed to facilitate the linking of common attributes between nodes in Maya. It's especially useful when you need to connect several attributes which share the same name in the placement and file nodes.

        Parameters:
            attr (str): The name of the attribute to connect between the nodes.
                        For example, 'coverage', 'translateFrame', 'rotateFrame', etc.
            place (str): The name of the placement node which has the attribute to connect.
            file (str): The name of the file node where the attribute will be connected to.

        Note:
            For attributes named differently between the place and file nodes, you should use the 'connectAttr' function with the respective attribute names.

        Example:
            connect_attributes('coverage', 'place2d', 'fileNode')
            connect_attributes('translateFrame', 'place2d', 'fileNode')
            connect_attributes('rotateFrame', 'place2d', 'fileNode')
            connect_attributes('mirror', 'place2d', 'fileNode')
            ...

            pm.connectAttr(f'{place}.outUV', f'{file}.uv', f=1)
            pm.connectAttr(f'{place}.outUvFilterSize', f'{file}.uvFilterSize', f=1)
        """
        pm.connectAttr("{}.{}".format(place, attr), "{}.{}".format(file, attr), f=1)

    @staticmethod
    def connect_multi_attr(*args, force=True):
        """Connect multiple node attributes at once.

        Parameters:
            args (tuple): Attributes as two element tuples. ie. (<connect from attribute>, <connect to attribute>)

        Example:
            connect_multi_attr(
                (node1.outColor, node2.aiSurfaceShader),
                (node1.outColor, node3.baseColor),
                (node4.outNormal, node5.normalCamera),
            )
        """
        for frm, to in args:
            try:
                pm.connectAttr(frm, to)
            except Exception as error:
                print("# Error:", __file__, error, "#")

    @staticmethod
    def create_assembly(nodes, assembly_name="assembly#", duplicate=False):
        """Create an assembly by parenting the input nodes to a new assembly node.

        Parameters:
            nodes (list): A list of nodes to include in the assembly.
            assembly_name (str, optional): The name of the assembly node. Defaults to 'assembly#'.
            duplicate (bool, optional): If True, duplicates the input nodes before parenting. Defaults to False.

        Returns:
            pm.PyNode: The assembly node with added properties:
                - addChild (function): Adds a new child to the assembly node.
                - children (function): Returns the list of children under the assembly node.
        """
        assembly_node = pm.assembly(name=assembly_name)

        for node in nodes:
            if duplicate:
                node = pm.duplicate(node)[0]
            pm.parent(node, assembly_node)

        assembly_node.addChild = lambda child: pm.parent(child, assembly_node)
        assembly_node.children = lambda: pm.listRelatives(assembly_node, children=True)

        return assembly_node

    @staticmethod
    def get_instances(objects=None, return_parent_objects=False):
        """get any intances of given object, or if None given; get all instanced objects in the scene.

        Parameters:
            objects (str/obj/list): Parent object/s.
            return_parent_objects (bool): Return instances and the given parent objects together.

        Returns:
            (list)
        """
        instances = []

        if objects is None:  # get all instanced objects in the scene.
            import maya.OpenMaya as om

            iterDag = om.MItDag(om.MItDag.kBreadthFirst)
            while not iterDag.isDone():
                instanced = om.MItDag.isInstanced(iterDag)
                if instanced:
                    instances.append(iterDag.fullPathName())
                iterDag.next()
        else:
            shapes = pm.listRelatives(objects, s=1)
            instances = pm.listRelatives(shapes, ap=1)
            if not return_parent_objects:
                [instances.remove(obj) for obj in objects]

        return instances

    @classmethod
    def convert_to_instances(cls, objects=[], append=""):
        """The first selected object will be instanced across all other selected objects.

        Parameters:
            objects (list): A list of objects to convert to instances. The first object will be the instance parent.
            append (str): Append a string to the end of any instanced objects. ie. '_INST'

        Returns:
            (list) The instanced objects.

        Example:
            convert_to_instances(pm.ls(sl=1))
        """
        pm.undoInfo(openChunk=True)
        # Get the world space obj pivot.
        p0x, p0y, p0z = pm.xform(objects[0], q=True, rotatePivot=1, worldSpace=1)
        # Get the obj pivot.
        # pivot = pm.xform(objects[0], q=True, rotatePivot=1, objectSpace=1)

        for obj in objects[1:]:
            name = obj.name()
            objParent = pm.listRelatives(obj, parent=1)

            instance = pm.instance(objects[0])

            cls.uninstance(obj)
            pm.makeIdentity(obj, apply=1, translate=1, rotate=0, scale=0)

            # Move object to center of the last selected items bounding box # pm.xform(instance, translation=pos, worldSpace=1, relative=1) #move to the original objects location.
            pm.matchTransform(instance, obj, position=1, rotation=1, scale=1, pivots=1)

            try:
                # Parent the instance under the original objects parent.
                pm.parent(instance, objParent)
            except RuntimeError:  # It is already a child of the parent.
                pass

            # Delete history for the object so that the namespace is cleared.
            pm.delete(obj, constructionHistory=True)
            pm.delete(obj)
            pm.rename(instance, name + append)
        pm.select(objects[1:])
        pm.undoInfo(closeChunk=True)

        return objects[1:]

    @classmethod
    def uninstance(cls, objects):
        """Un-Instance the given objects.

        Parameters:
            objects (str/obj/list): The objects to un-instance. If 'all' is given all instanced objects in the scene will be uninstanced.
        """
        if objects == "all":
            objects = cls.get_instances()

        for obj in pm.ls(objects):
            children = pm.listRelatives(obj, fullPath=1, children=1)
            parents = pm.listRelatives(children[0], fullPath=1, allParents=1)

            if len(parents) > 1:
                duplicatedObject = pm.duplicate(obj)
                pm.delete(obj)
                pm.rename(duplicatedObject[0], obj)


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    pass

# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------



# !/usr/bin/python
# coding=utf-8
import os

try:
    import pymel.core as pm
except ImportError as error:
    print(__file__, error)
import pythontk as ptk


class Project:
    """ """

    @staticmethod
    def get_recent_files(index=None, format="standard"):
        """
        Get a list of recent files.

        Parameters:
            index (slice or int): Return the recent file directory path at the given index or slice.
                    Index 0 would be the most recent file.
                    For example, use index=slice(0, 5) to get the 5 most recent files.
                    If there are only 3 files, it will return those 3 files without throwing an error.
            format (str): Defines the format of the returned paths. Possible options are 'standard', 'timestamp',
                    'standard|timestamp', 'timestamp|standard'. 'standard' returns paths as strings, 'timestamp'
                    returns timestamped paths, 'standard|timestamp' returns a dictionary with standard paths as
                    keys and timestamped paths as values, 'timestamp|standard' does the opposite.

        Returns:
            (list or dict): A list or dictionary of recent files depending on the 'format' parameter.

        Examples:
            get_recent_files() --> Returns all recent files in standard format
            get_recent_files(0) --> Returns the most recent file in standard format
            get_recent_files(slice(0, 5)) --> Returns the 5 most recent files in standard format.
            get_recent_files(format='timestamp') --> Returns all recent files in timestamp format.
            get_recent_files(format='standard|timestamp') --> Returns a dictionary with standard paths as keys and timestamped paths as values.
        """
        files = pm.optionVar(query="RecentFilesList")
        if not files:
            return []

        result = [
            ptk.format_path(f)
            for f in reversed(files)
            if ptk.is_valid(f) and "Autosave" not in f
        ]

        if index is not None:
            try:
                result = result[index]
            except (IndexError, TypeError):
                print(f"Incorrect index or slice. Returning empty list.")
                return []

        format = format.split("|")
        if len(format) == 2 and "timestamp" in format and "standard" in format:
            if format[0] == "timestamp":
                result = {ptk.time_stamp(res): res for res in result}
            else:
                result = {res: ptk.time_stamp(res) for res in result}
        elif "timestamp" in format:
            result = [ptk.time_stamp(res) for res in result]
        # else return the standard format

        return result

    @staticmethod
    def get_recent_projects(index=None, format="standard"):
        """
        Get a list of recently set projects.

        Parameters:
            index (slice or int): Return the recent project directory path at the given index or slice.
                    Index 0 would be the most recent project.
                    For example, use index=slice(0, 5) to get the 5 most recent projects.
                    If there are only 3 projects, it will return those 3 projects without throwing an error.
            format (str): Defines the format of the returned paths. Possible options are 'standard', 'timestamp',
                    'standard|timestamp', 'timestamp|standard'. 'standard' returns paths as strings, 'timestamp'
                    returns timestamped paths, 'standard|timestamp' returns a dictionary with standard paths as
                    keys and timestamped paths as values, 'timestamp|standard' does the opposite.

        Returns:
            (list or dict): A list or dictionary of recent projects depending on the 'format' parameter.

        Examples:
            get_recent_projects() --> Returns all recent projects in standard format
            get_recent_projects(0) --> Returns the most recent project in standard format
            get_recent_projects(slice(0, 5)) --> Returns the 5 most recent projects in standard format.
            get_recent_projects(format='timestamp') --> Returns all recent projects in timestamp format.
            get_recent_projects(format='standard|timestamp') --> Returns a dictionary with standard paths as keys and timestamped paths as values.
        """
        files = pm.optionVar(query="RecentProjectsList")
        if not files:
            return []

        result = [ptk.format_path(f) for f in reversed(files) if ptk.is_valid(f)]

        if index is not None:
            try:
                result = result[index]
            except (IndexError, TypeError):
                print("Incorrect index or slice. Returning empty list.")
                return []

        format = format.split("|")
        if len(format) == 2 and "timestamp" in format and "standard" in format:
            if format[0] == "timestamp":
                result = {ptk.time_stamp(res): res for res in result}
            else:
                result = {res: ptk.time_stamp(res) for res in result}
        elif "timestamp" in format:
            result = [ptk.time_stamp(res) for res in result]
        # else return the standard format

        return result

    @staticmethod
    def get_recent_autosave(index=None, format="standard"):
        """
        Returns a list of recent Maya autosave files (.mb and .ma), sorted by timestamp.

        Parameters:
            index (slice or int): Return the recent autosave file directory path at the given index or slice.
                    Index 0 would be the most recent autosave file.
                    For example, use index=slice(0, 5) to get the 5 most recent autosave files.
                    If there are only 3 autosave files, it will return those 3 autosave files without throwing an error.
            format (str): Defines the format of the returned paths. Possible options are 'standard', 'timestamp',
                    'standard|timestamp', 'timestamp|standard'. 'standard' returns paths as strings, 'timestamp'
                    returns timestamped paths, 'standard|timestamp' returns a dictionary with standard paths as
                    keys and timestamped paths as values, 'timestamp|standard' does the opposite.

        Returns:
            (list or dict): A list or dictionary of recent autosave files depending on the 'format' parameter.

        Examples:
            get_recent_autosave() --> Returns all recent autosave files in standard format
            get_recent_autosave(0) --> Returns the most recent autosave file in standard format
            get_recent_autosave(slice(0, 5)) --> Returns the 5 most recent autosave files in standard format.
            get_recent_autosave(format='timestamp') --> Returns all recent autosave files in timestamp format.
            get_recent_autosave(format='standard|timestamp') --> Returns a dictionary with standard paths as keys and timestamped paths as values.
        """
        import glob
        import itertools

        autosave_dirs = [str(pm.workspace(q=True, rd=1)) + "autosave"]
        env_autosave_dir = os.environ.get("MAYA_AUTOSAVE_FOLDER")
        if env_autosave_dir is not None:
            autosave_dirs += env_autosave_dir.split(";")
        autosave_dirs.append(os.path.expanduser("~/maya/autosave"))

        result = []
        for autosave_dir in autosave_dirs:
            if not os.path.exists(autosave_dir):
                continue

        files = itertools.chain(
            glob.iglob(os.path.join(autosave_dir, "*.mb")),
            glob.iglob(os.path.join(autosave_dir, "*.ma")),
        )

        for file in files:
            result.append(ptk.format_path(file))

        if index is not None:
            try:
                result = result[index]
            except (IndexError, TypeError):
                print("Incorrect index or slice. Returning empty list.")
                return []

        format = format.split("|")
        if len(format) == 2 and "timestamp" in format and "standard" in format:
            if format[0] == "timestamp":
                result = {ptk.time_stamp(res): res for res in result}
            else:
                result = {res: ptk.time_stamp(res) for res in result}
        elif "timestamp" in format:
            result = [ptk.time_stamp(res) for res in result]
        # else return the standard format

        return result

    @staticmethod
    def get_workspace_scenes(fullPath=True):
        """Get a list of maya scene files from the current workspace directory.

        Parameters:
            fullPath (bool): Return the full path instead of just the filename.

        Returns:
            (list)
        """
        workspace_dir = str(pm.workspace(q=True, rd=1))  # get current project path.

        files = ptk.get_dir_contents(
            workspace_dir, "filepath", inc_files=("*.mb", "*.ma")
        )
        # Replace any backslashes with forward slashes.
        result = [ptk.format_path(f) for f in files]

        if not fullPath:
            result = [f.split("\\")[-1] for f in result]

        return result

    @staticmethod
    def reference_scene(file_path):
        """Reference a Maya scene.

        Parameters:
            file_path (str): The path to the Maya scene file to reference.
        """
        if os.path.exists(file_path):
            pm.system.createReference(file_path)
        else:
            raise FileNotFoundError(f"No such file: '{file_path}'")

    @staticmethod
    def remove_reference(file_path):
        """Remove a reference to a Maya scene.

        Parameters:
            file_path (str): The path to the Maya scene file to remove the reference to.
        """
        ref_node = pm.system.FileReference(file_path)
        if ref_node.isReferenced():
            ref_node.remove()

    @staticmethod
    def is_referenced(file_path):
        """Check if a Maya scene is referenced.

        Parameters:
            file_path (str): The path to the Maya scene file to check.

        Returns:
            (bool): True if the scene is referenced, False otherwise.
        """
        ref_node = pm.system.FileReference(file_path)
        return ref_node.isReferenced()

    @staticmethod
    def get_reference_nodes(file_path):
        """Get the nodes from a referenced Maya scene.

        Parameters:
            file_path (str): The path to the Maya scene file to get the nodes from.

        Returns:
            (list): A list of nodes in the referenced scene.
        """
        ref_node = pm.system.FileReference(file_path)
        if ref_node.isReferenced():
            return ref_node.nodes()
        else:
            return []

    @staticmethod
    def list_references():
        """List all references in the current Maya scene.

        Returns:
            (list): A list of all references in the current Maya scene.
        """
        return [ref.filePath() for ref in pm.system.listReferences()]


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    pass

# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------



# !/usr/bin/python
# coding=utf-8
try:
    import pymel.core as pm
except ImportError as error:
    print(__file__, error)
import pythontk as ptk

# from this package:
from mayatk import node_utils


class RigUtils(object):
    """ """

    @staticmethod
    def create_locator(name=None, pos=(), scale=1):
        """Create a locator with the given scale.

        Parameters:
            name (str): Name the locator.
            pos (tuple): The desired location in world space.
            scale (float) = The desired scale of the locator.

        Returns:
            (obj) locator
        """
        loc = pm.spaceLocator()

        if name:
            pm.rename(loc, name)
        if pos:
            pm.move(pos[0], pos[1], pos[2], loc)
        if scale != 1:
            pm.scale(loc, scale, scale, scale)  # scale the locator

        return loc

    @classmethod
    def remove_locator(cls, objects):
        """Remove a parented locator from the child object.

        Parameters:
            obj (str/obj/list): The child object or the locator itself.
        """
        pm.undoInfo(openChunk=1)

        for obj in pm.ls(objects, long=True, objectsOnly=True):
            if not pm.objExists(obj):
                continue

            elif (
                node_utils.NodeUtils.is_locator(obj)
                and not node_utils.NodeUtils.get_type(obj)
                and not node_utils.NodeUtils.get_children(obj)
            ):
                pm.delete(obj)
                continue

            # unlock attributes
            cls.set_attr_lock_state(obj, translate=False, rotate=False, scale=False)

            if not node_utils.NodeUtils.is_locator(obj):
                try:  # if the 'obj' is not a locator, check if it's parent is.
                    obj = node_utils.NodeUtils.get_parent(obj)
                    if not node_utils.NodeUtils.is_locator(obj):
                        pm.inViewMessage(
                            status_message="Error: Unable to remove locator for the given object.",
                            pos="topCenter",
                            fade=True,
                        )
                        continue
                except IndexError:
                    pm.inViewMessage(
                        status_message="Error: Unable to remove locator for the given object.",
                        pos="topCenter",
                        fade=True,
                    )
                    continue

            # unparent child object
            children = node_utils.NodeUtils.get_children(obj)
            for child in children:
                pm.parent(child, world=True)

            # remove locator
            pm.delete(obj)

        pm.undoInfo(closeChunk=1)

    @staticmethod
    def reset_pivot_transforms(objects):
        """Reset Pivot Transforms"""
        objs = pm.ls(type=("transform", "geometryShape"), sl=1)

        if len(objs) > 0:
            pm.xform(cp=1)

        pm.manipPivot(ro=1, rp=1)

    @staticmethod
    def bake_custom_pivot(objects, position=False, orientation=False):
        """ """
        transforms = pm.ls(objects, transforms=1)
        shapes = pm.ls(objects, shapes=1)
        objects = transforms + pm.listRelatives(
            shapes, path=1, parent=1, type="transform"
        )

        ctx = pm.currentCtx()
        pivotModeActive = 0
        customModeActive = 0
        if ctx in ("RotateSuperContext", "manipRotateContext"):  # Rotate tool
            customOri = pm.manipRotateContext("Rotate", q=1, orientAxes=1)
            pivotModeActive = pm.manipRotateContext("Rotate", q=1, editPivotMode=1)
            customModeActive = pm.manipRotateContext("Rotate", q=1, mode=1) == 3
        elif ctx in ("scaleSuperContext", "manipScaleContext"):  # Scale tool
            customOri = pm.manipScaleContext("Scale", q=1, orientAxes=1)
            pivotModeActive = pm.manipScaleContext("Scale", q=1, editPivotMode=1)
            customModeActive = pm.manipScaleContext("Scale", q=1, mode=1) == 6
        else:  # use the move tool orientation
            customOri = pm.manipMoveContext(
                "Move", q=1, orientAxes=1
            )  # get custom orientation
            pivotModeActive = pm.manipMoveContext("Move", q=1, editPivotMode=1)
            customModeActive = pm.manipMoveContext("Move", q=1, mode=1) == 6
            if ctx not in ("moveSuperContext", "manipMoveContext"):  # Move tool
                otherToolActive = 1  # some other tool

        if orientation and customModeActive:
            if not position:
                pm.mel.error(
                    (pm.mel.uiRes("m_bakeCustomToolPivot.kWrongAxisOriToolError"))
                )
                return

            from math import degrees

            cX, cY, cZ = customOri = [
                degrees(customOri[0]),
                degrees(customOri[1]),
                degrees(customOri[2]),
            ]

            pm.rotate(
                objects, cX, cY, cZ, a=1, pcp=1, pgp=1, ws=1, fo=1
            )  # Set object(s) rotation to the custom one (preserving child transform positions and geometry positions)

        if position:
            for obj in objects:
                # Get pivot in parent space
                m = pm.xform(obj, q=1, m=1)
                p = pm.xform(obj, q=1, os=1, sp=1)
                oldX, oldY, oldZ = old = [
                    (p[0] * m[0] + p[1] * m[4] + p[2] * m[8] + m[12]),
                    (p[0] * m[1] + p[1] * m[5] + p[2] * m[9] + m[13]),
                    (p[0] * m[2] + p[1] * m[6] + p[2] * m[10] + m[14]),
                ]

                pm.xform(obj, zeroTransformPivots=1)  # Zero out pivots

                # Translate obj(s) back to previous pivot (preserving child transform positions and geometry positions)
                newX, newY, newZ = new = pm.getAttr(
                    obj.name() + ".translate"
                )  # obj.translate
                pm.move(
                    obj, oldX - newX, oldY - newY, oldZ - newZ, pcp=1, pgp=1, ls=1, r=1
                )

        if pivotModeActive:
            pm.ctxEditMode()  # exit pivot mode

        # Set the axis orientation mode back to obj
        if orientation and customModeActive:
            if ctx in ("RotateSuperContext", "manipRotateContext"):
                pm.manipPivot(rotateToolOri=0)
            elif ctx in ("scaleSuperContext", "manipScaleContext"):
                pm.manipPivot(scaleToolOri=0)
            else:  # Some other tool #Set move tool to obj mode and clear the custom ori. (so the tool won't restore it when activated)
                pm.manipPivot(moveToolOri=0)
                if not ctx in ("moveSuperContext", "manipMoveContext"):
                    pm.manipPivot(ro=1)

    @classmethod
    def set_attr_lock_state(
        cls, objects, translate=None, rotate=None, scale=None, **kwargs
    ):
        """Lock/Unlock any attribute for the given objects, by passing it into kwargs as <attr>=<value>.
        A 'True' value locks the attribute, 'False' unlocks, while 'None' leaves the state unchanged.

        Parameters:
            objects (str/obj/list): The object(s) to lock/unlock attributes of.
            translate (bool): Lock/Unlock all translate x,y,z values at once.
            rotate (bool): Lock/Unlock all rotate x,y,z values at once.
            scale (bool): Lock/Unlock all scale x,y,z values at once.
        """
        objects = pm.ls(objects, transforms=1, long=True)

        attrs_and_state = {
            (
                "tx",
                "ty",
                "tz",
            ): translate,  # attributes and state. ex. ('tx','ty','tz'):False
            ("rx", "ry", "rz"): rotate,
            ("sx", "sy", "sz"): scale,
        }

        attrs_and_state.update(kwargs)  # update the dict with any values from kwargs.

        for obj in objects:
            try:
                if node_utils.NodeUtils.is_locator(obj):
                    obj = pm.listRelatives(obj, children=1, type="transform")[0]
            except IndexError:
                return

            for attrs, state in attrs_and_state.items():
                if state is None:
                    continue
                for a in ptk.make_iterable(attrs):
                    pm.setAttr("{}.{}".format(obj, a), lock=state)

    @staticmethod
    def create_group(
        objects=[],
        name="",
        zero_translation=False,
        zero_rotation=False,
        zero_scale=False,
    ):
        """Create a group containing any given objects.

        Parameters:
            objects (str/obj/list): The object(s) to group.
            name (str): Name the group.
            zero_translation (bool): Freeze translation before parenting.
            zero_rotation (bool): Freeze rotation before parenting.
            zero_scale (bool): Freeze scale before parenting.

        Returns:
            (obj) the group.
        """
        grp = pm.group(empty=True, n=name)
        try:
            pm.parent(grp, objects)
        except Exception as error:
            print(
                f"{__file__} in create_group\n\t# Error: Unable to parent object(s): {error} #"
            )

        if zero_translation:
            for attr in ("tx", "ty", "tz"):
                pm.setAttr(getattr(grp, attr), 0)  # pm.setAttr(node.translate, 0)
        if zero_rotation:
            for attr in ("rx", "ry", "rz"):
                pm.setAttr(getattr(grp, attr), 0)
        if zero_scale:
            for attr in ("sx", "sy", "sz"):
                pm.setAttr(getattr(grp, attr), 0)

        pm.parent(grp, world=True)
        return grp

    @classmethod
    def create_group_with_first_obj_lra(cls, objects, name="", freeze_transforms=True):
        """Creates a group using the first object to define the local rotation axis.

        Parameters:
            objects (str/obj/list): The objects to group. The first object will be used to define the groups LRA.
            name (str): The group name.
            freeze_transforms (bool): Freeze transforms on group child objects.
        """
        try:
            obj, *other = pm.ls(objects, transforms=1)
        except IndexError:
            print(
                f"{__file__} in create_group_with_first_obj_lra\n\t# Error: Operation requires at least one object. #"
            )
            return None

        pm.undoInfo(openChunk=1)
        cls.bake_custom_pivot(
            obj, position=True, orientation=True
        )  # pm.mel.BakeCustomPivot(obj) #bake the pivot on the object that will define the LRA.

        grp = pm.group(empty=True)
        pm.parent(grp, obj)

        pm.setAttr(grp.translate, (0, 0, 0))
        pm.setAttr(grp.rotate, (0, 0, 0))

        objParent = pm.listRelatives(obj, parent=1)
        pm.parent(
            grp, objParent
        )  # parent the instance under the original objects parent.

        try:
            pm.parent(obj, grp)
        except Exception:  # root level objects
            pm.parent(grp, world=True)
            pm.parent(obj, grp)

        for o in other:  # parent any other objects to the new group.
            pm.parent(o, grp)
            if freeze_transforms:
                pm.makeIdentity(o, apply=True)  # freeze transforms on child objects.

        if not name and objParent:  # name the group.
            pm.rename(grp, objParent[0].name())
        elif not name:
            pm.rename(grp, obj.name())
        else:
            pm.rename(grp, name)
        pm.undoInfo(closeChunk=1)

        return grp

    @classmethod
    def create_locator_at_object(
        cls,
        objects,
        parent=False,
        freeze_transforms=False,
        bake_child_pivot=False,
        grp_suffix="_GRP#",
        loc_suffix="_LOC#",
        obj_suffix="_GEO#",
        strip_digits=False,
        strip_suffix=False,
        scale=1,
        lock_translate=False,
        lock_rotation=False,
        lock_scale=False,
    ):
        """Create locators with the same transforms as any selected object(s).
        If there are vertices selected it will create a locator at the center of the selected vertices bounding box.

        Parameters:
            objects (str/obj/list): A list of objects, or an object name to create locators at.
            parent (bool): Parent the object to the locator. (default=False)
            freeze_transforms (bool): Freeze transforms on the locator. (default=True)
            bake_child_pivot (bool): Bake pivot positions on the child object. (default=True)
            grp_suffix (str): A string appended to the end of the created groups name. (default: '_GRP#')
            loc_suffix (str): A string appended to the end of the created locators name. (default: '_LOC#')
            obj_suffix (str): A string appended to the end of the existing objects name. (default: '_GEO#')
            strip_digits (bool): Strip numeric characters from the string. If the resulting name is not unique, maya will append a trailing digit. (default=False)
            strip_suffix (str): Strip any existing suffix. A suffix is defined by the last '_' (if one exists) and any chars trailing. (default=False)
            scale (float) = The scale of the locator. (default=1)
            lock_translate (bool): Lock the translate values of the child object. (default=False)
            lock_rotation (bool): Lock the rotation values of the child object. (default=False)
            lock_scale (bool): Lock the scale values of the child object. (default=False)
            remove (bool): Removes the locator and any child locks. (not valid with component selections) (default=False)

        Example:
            createLocatorAtSelection(strip='_GEO', suffix='', strip_digits=True, parent=True, lock_translate=True, lock_rotation=True)
        """
        getSuffix = (
            lambda o: loc_suffix
            if node_utils.NodeUtils.is_locator(o)
            else grp_suffix
            if node_utils.NodeUtils.is_group(o)
            else obj_suffix
        )  # match the correct suffix to the object type.

        pm.undoInfo(openChunk=1)

        for obj in pm.ls(objects, long=True, type="transform"):
            if bake_child_pivot:
                cls.bake_custom_pivot(obj, position=1, orientation=1)

            vertices = pm.filterExpand(obj, sm=31)  # returns a string list.
            if vertices:
                objName = vertices[0].split(".")[0]
                obj = pm.ls(objName)

                loc = cls.create_locator(scale=scale)

                xmin, ymin, zmin, xmax, ymax, zmax = pm.exactWorldBoundingBox(vertices)
                x, y, z = pos = (
                    (xmin + xmax) / 2,
                    (ymin + ymax) / 2,
                    (zmin + zmax) / 2,
                )
                pm.move(x, y, z, loc)

            else:  # object:
                loc = cls.create_locator(scale=scale)
                tempConst = pm.parentConstraint(obj, loc, maintainOffset=False)
                pm.delete(tempConst)

            try:
                if parent:
                    origParent = pm.listRelatives(obj, parent=1)

                    grp = cls.create_group(obj, zero_translation=1, zero_rotation=1)
                    pm.rename(
                        grp,
                        ptk.format_suffix(
                            obj.name(),
                            suffix=getSuffix(grp),
                            strip=(obj_suffix, grp_suffix, loc_suffix),
                            strip_trailing_ints=strip_digits,
                            strip_trailing_alpha=strip_suffix,
                        ),
                    )

                    pm.parent(obj, loc)
                    pm.parent(loc, grp)
                    pm.parent(grp, origParent)

                if freeze_transforms:  # freeze transforms before baking pivot.
                    cls.set_attr_lock_state(
                        obj, translate=False, rotate=False, scale=False
                    )  # assure attributes are unlocked.
                    pm.makeIdentity(obj, apply=True, normal=1)
                    pm.makeIdentity(
                        loc, apply=True, normal=1
                    )  # 1=the normals on polygonal objects will be frozen. 2=the normals on polygonal objects will be frozen only if its a non-rigid transformation matrix.

                pm.rename(
                    loc,
                    ptk.format_suffix(
                        obj.name(),
                        suffix=getSuffix(loc),
                        strip=(obj_suffix, grp_suffix, loc_suffix),
                        strip_trailing_ints=strip_digits,
                        strip_trailing_alpha=strip_suffix,
                    ),
                )
                pm.rename(
                    obj,
                    ptk.format_suffix(
                        obj.name(),
                        suffix=getSuffix(obj),
                        strip=(obj_suffix, grp_suffix, loc_suffix),
                        strip_trailing_ints=strip_digits,
                        strip_trailing_alpha=strip_suffix,
                    ),
                )

                cls.set_attr_lock_state(
                    obj,
                    translate=lock_translate,
                    rotate=lock_rotation,
                    scale=lock_scale,
                )

            except Exception as error:
                pm.delete(loc)
                raise (error)

        pm.undoInfo(closeChunk=1)


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    sel = pm.ls(sl=1)

    RigUtils.create_locator_at_object(
        sel,
        parent=1,
        freeze_transforms=1,
        bake_child_pivot=1,
        grp_suffix="_GRP",
        loc_suffix="_LCTR",
        obj_suffix="_GEO",
        strip_digits=1,
        strip_suffix=1,
        scale=1,
        lock_translate=0,
        lock_rotation=0,
        lock_scale=0,
    )

if __name__ == "__main__":
    pass

# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------


# deprecated ---------------------



# !/usr/bin/python
# coding=utf-8
try:
    import pymel.core as pm
except ImportError as error:
    print(__file__, error)
import pythontk as ptk

# from this package:
from mayatk import core_utils, node_utils


class GetComponentsMixin:
    """ """

    componentTypes = [  # abv, singular, plural, full, int, hex
        (
            "vtx",
            "vertex",
            "vertices",
            "Polygon Vertex",
            31,
            0x0001,
        ),
        (
            "e",
            "edge",
            "edges",
            "Polygon Edge",
            32,
            0x8000,
        ),
        (
            "f",
            "face",
            "faces",
            "Polygon Face",
            34,
            0x0008,
        ),
        (
            "uv",
            "texture",
            "texture coordinates",
            "Polygon UV",
            35,
            0x0010,
        ),
        (
            "cv",
            "control vertex",
            "control vertices",
            "Control Vertex",
            28,
            None,
        ),
        (
            "vtxf",
            "vertexFace",
            "vertexFaces",
            "Polygon Vertex Face",
            70,
            None,
        ),
        (
            None,
            "edit point",
            "edit points",
            "Edit Point",
            30,
            None,
        ),
        (
            None,
            "handle",
            "handles",
            "Handle",
            0,
            None,
        ),
        (
            None,
            "nurbs surface",
            "nurbs surfaces",
            "Nurbs Curves On Surface",
            11,
            None,
        ),
        (
            None,
            "subd mesh point",
            "subd mesh points",
            "Subdivision Mesh Point",
            36,
            None,
        ),
        (
            None,
            "subd mesh edge",
            "subd mesh edges",
            "Subdivision Mesh Edge",
            37,
            None,
        ),
        (
            None,
            "subd mesh face",
            "subd mesh faces",
            "Subdivision Mesh Face",
            38,
            None,
        ),
        (
            None,
            "curve parameter point",
            "curve parameter points",
            "Curve Parameter Point",
            39,
            None,
        ),
        (
            None,
            "curve knot",
            "curve knots",
            "Curve Knot",
            40,
            None,
        ),
        (
            None,
            "surface parameter point",
            "surface parameter points",
            "Surface Parameter Point",
            41,
            None,
        ),
        (
            None,
            "surface knot",
            "surface knots",
            "Surface Knot",
            42,
            None,
        ),
        (
            None,
            "surface range",
            "surface ranges",
            "Surface Range",
            43,
            None,
        ),
        (
            None,
            "trim surface edge",
            "trim surface edges",
            "Trim Surface Edge",
            44,
            None,
        ),
        (
            None,
            "surface isoparm",
            "surface isoparms",
            "Surface Isoparm",
            45,
            None,
        ),
        (
            None,
            "lattice point",
            "lattice points",
            "Lattice Point",
            46,
            None,
        ),
        (
            None,
            "particle",
            "particles",
            "Particle",
            47,
            None,
        ),
        (
            None,
            "scale pivot",
            "scale pivots",
            "Scale Pivot",
            49,
            None,
        ),
        (
            None,
            "rotate pivot",
            "rotate pivots",
            "Rotate Pivot",
            50,
            None,
        ),
        (
            None,
            "select handle",
            "select handles",
            "Select Handle",
            51,
            None,
        ),
        (
            None,
            "nurbs surface face",
            "nurbs surface faces",
            "NURBS Surface Face",
            72,
            None,
        ),
        (
            None,
            "subd mesh UV",
            "subd mesh UVs",
            "Subdivision Mesh UV",
            73,
            None,
        ),
    ]

    @classmethod
    def get_component_type(cls, component, returned_type="abv"):
        """Get the type of a given component.

        Parameters:
            obj (str/obj/list): A single maya component.
                    If multiple components are given, only the first will be sampled.
            returned_type (str): Specify the desired return value type. (default: 'str')
                    (valid: 'full' - object type as a string.
                                    'int' - maya mask value as an integer.
                                    'hex' - hex value. ie. 0x0001
                                    'abv' - abreviated object type as a string. ie. 'vtx'
        Returns:
            (str)(int) dependant on 'returned_type' arg.

        Example:
            get_component_type('cyl.e[:]') #returns: 'e'
            get_component_type('cyl.vtx[:]', 'abv') #returns: 'vtx'
            get_component_type('cyl.e[:]', 'int') #returns: 32
        """
        for a, s, p, f, i, h in cls.componentTypes:
            try:
                if pm.filterExpand(component, sm=i):
                    if returned_type == "abv":
                        return a
                    elif returned_type == "full":
                        return f
                    elif returned_type == "int":
                        return i
                    elif returned_type == "hex":
                        return h
                    elif returned_type == "plural":
                        return p
                    else:
                        return s
            except Exception as e:
                print(
                    f'File "{__file__}" in get_component_type\n# Error: Not a valid component. #\n{e}{"(empty string)" if component == "" else ""}',
                )
                break
        return None

    @classmethod
    def convert_alias(cls, component_type, returned_type="abv"):
        """Return an alternate component alias for the given alias.
        ie. a hex value of 0x0001 for 'vertex'
        If nothing is found, a value of 'None' will be returned.

        Parameters:
            component_type () = A component type. ex. 'vertex', 'vtx', 31, or 0x0001
            returned_type (str): The desired returned alias.  (default: 'abv')
                    (valid: 'abv', 'singular', 'plural', 'str', 'int', 'hex')

        Returns:
            (str)(int)(hex)(None) dependant on 'returned_type' argument.

        Example:
            convert_alias('vertex', 'hex') #returns: 0x0001
            convert_alias(0x0001, 'str') #returns: 'Polygon Vertex'
        """
        rtypes = ("abv", "singular", "plural", "full", "int", "hex")

        for t in cls.componentTypes:
            if component_type in t:
                index = rtypes.index(returned_type)
                return t[index]
        return None

    @classmethod
    def convert_component_type(
        cls, components, component_type, returned_type="str", flatten=False
    ):
        """Convert component(s) to it's sub-components of the given type.

        Parameters:
            components (str/obj/list): The components(s) to convert.
            component_type (str): The desired returned component type.
                    valid: 'vtx' (or 'vertex', 'vertices', 'Polygon Vertex', 31, 0x0001),
                            and the same for each: 'edge', 'uv', 'face'.
            returned_type (str): The desired returned object type.
                    (valid: 'str'(default), 'obj'(shape object), 'transform'(as string), 'int'(valid only at sub-object level).
            flatten (bool): Flattens the returned list of objects so that each component is it's own element.

        Returns:
            (list)(dict)

        Example:
            convert_component_type('obj.vtx[:2]', 'vertex') #returns: ['obj.vtx[0:2]']
            convert_component_type('obj.vtx[:2]', 'face') #returns: ['obj.f[0:2]', 'obj.f[11:14]', 'obj.f[23]']
            convert_component_type('obj.vtx[:2]', 'edge') #returns: ['obj.e[0:2]', 'obj.e[11]', 'obj.e[24:26]', 'obj.e[36:38]']
            convert_component_type('obj.vtx[:2]', 'uv') #returns: ['obj.map[0:2]', 'obj.map[12:14]', 'obj.map[24]']
        """
        d = {
            "vtx": "toVertex",
            "e": "toEdge",
            "uv": "toUV",
            "f": "toFace",
            "uv": "toUV",
            "shell": "toShell",
            "vertexFace": "toVertexFace",
        }
        typ = cls.convert_alias(
            component_type
        )  # get the correct component_type variable from possible args.

        if typ not in d:
            return components
        components = pm.polyListComponentConversion(
            components, **{d[typ.lower()]: True}
        )
        return core_utils.CoreUtils.convert_array_type(
            components, returned_type=returned_type, flatten=flatten
        )

    @classmethod
    def convert_int_to_component(
        cls, obj, integers, component_type, returned_type="str", flatten=False
    ):
        """Convert the given integers to components of the given object.

        Parameters:
            obj (str/obj/list): The object to convert to vertices of.
            integers (list): The integer(s) to convert.
            component_type (str): The desired returned component type.
                    valid: 'vtx' (or 'vertex', 'vertices', 'Polygon Vertex', 31, 0x0001),
                            and the same for each: 'edge', 'uv', 'face'.
            returned_type (str): The desired returned object type.
                    (valid: 'str'(default), 'obj'(shape object), 'transform'(as string), 'int'(valid only at sub-object level).
            flatten (bool): Flattens the returned list of objects so that each component is it's own element.

        Returns:
            (list)

        Example:
            convert_int_to_component('cyl', range(4), 'f') #returns: ['cylShape.f[0:3]']
        """
        obj = pm.ls(obj, objectsOnly=True)[0]
        objName = obj.name()

        def n(c):
            return "{}:{}".format(c[0], c[-1]) if len(c) > 1 else str(c[0])

        if not flatten:
            result = [
                "{}.{}[{}]".format(objName, component_type, n(c))
                for c in ptk.split_list(integers, "range")
            ]
        else:
            result = ["{}.{}[{}]".format(objName, component_type, c) for c in integers]

        return core_utils.CoreUtils.convert_array_type(
            result, returned_type=returned_type, flatten=flatten
        )

    @classmethod
    def filter_components(cls, components, inc=[], exc=[], flatten=False):
        """Filter the given components.

        Parameters:
            components (str/obj/list): The components(s) to filter.
            inc (str)(int)(obj/list): The component(s) to include.
            exc (str)(int)(obj/list): The component(s) to exclude.
                                    (exlude take precidence over include)
            flatten (bool): Flattens the returned list of objects so that each component is it's own element.

        Returns:
            (list)

        Example:
            filter_components('cyl.vtx[:]', 'cyl.vtx[:2]', 'cyl.vtx[1:23]') #returns: ['cyl.vtx[0]']
            filter_components('cyl.f[:]', range(2), range(1, 23)) #returns: ['cyl.f[0]']
        """
        typ = cls.get_component_type(components)
        etyp = core_utils.CoreUtils.get_array_type(components)
        etyp_inc = core_utils.CoreUtils.get_array_type(inc)
        etyp_exc = core_utils.CoreUtils.get_array_type(exc)

        if etyp_inc == "int" or etyp_exc == "int":
            try:
                obj = pm.ls(components, objectsOnly=True)[0]
            except IndexError as e:
                print(
                    f'File "{__file__}" in filter_components\n# Error: Operation requires at least one component. #\n{e}',
                )
                return []

        if etyp_inc == "int":
            inc = cls.convert_int_to_component(obj, inc, typ)
        inc = pm.ls(inc, flatten=True)

        if etyp_exc == "int":
            exc = cls.convert_int_to_component(obj, exc, typ)
        exc = pm.ls(exc, flatten=True)

        components = pm.ls(components, flatten=True)

        filtered = ptk.filter_list(components, inc=inc, exc=exc)
        result = core_utils.CoreUtils.convert_array_type(
            filtered, returned_type=etyp, flatten=flatten
        )
        return result

    @classmethod
    def get_components(
        cls,
        objects,
        component_type,
        returned_type="str",
        inc=[],
        exc=[],
        randomize=0,
        flatten=False,
    ):
        """Get the components of the given type from the given object(s).
        If no objects are given the current selection will be used.

        Parameters:
            objects (str/obj/list): The object(s) to get the components of. (Polygon, Polygon components)(default: current selection)
            component_type (str)(int): The component type to return. (valid: any type allowed in the 'convert_alias' method)
            returned_type (str): The desired returned object type.
                    (valid: 'str'(default), 'obj'(shape object), 'transform'(as string), 'int'(valid only at sub-object level).
            inc (str/obj/list): The component(s) to include.
            exc (str/obj/list): The component(s) to exclude. (exlude take precidence over include)
            randomize (float) = If a 0.1-1 value is given, random components will be returned with a quantity determined by the given ratio.
                                                    A value of 0.5 will return a 50% of the components of an object in random order.
            flatten (bool): Flattens the returned list of objects so that each component is it's own element.

        Returns:
            (list)(dict) Dependant on flags.

        Example:
            get_components(obj, 'vertex', 'str', '', 'obj.vtx[2:23]') #returns: ['objShape.vtx[0]', 'objShape.vtx[1]', 'objShape.vtx[24]', 'objShape.vtx[25]']
            get_components(obj, 'vertex', 'obj', '', 'obj.vtx[:23]') #returns: [MeshVertex('objShape.vtx[24]'), MeshVertex('objShape.vtx[25]')]
            get_components(obj, 'f', 'int') #returns: {nt.Mesh('objShape'): [(0, 35)]}
            get_components(obj, 'edges') #returns: ['objShape.e[0:59]']
            get_components(obj, 'edges', 'str', 'obj.e[:2]') #returns: ['objShape.e[0]', 'objShape.e[1]', 'objShape.e[2]']
        """
        components = cls.convert_component_type(objects, component_type)

        if inc or exc:
            components = cls.filter_components(components, inc=inc, exc=exc)

        if randomize:
            components = randomize(pm.ls(components, flatten=1), randomize)

        result = core_utils.CoreUtils.convert_array_type(
            components, returned_type=returned_type, flatten=flatten
        )
        return result


class Components(GetComponentsMixin):
    """ """

    @staticmethod
    def map_components_to_objects(components_list):
        """Maps a list of components to their respective objects.

        This function takes in a list of PyMel component objects and returns a
        dictionary where the keys are the names of the parent objects of the
        components, and the values are lists of components belonging to each object.

        Parameters:
            components_list (list): A list of PyMel component objects.

        Returns:
            dict: A dictionary mapping object names to lists of components.
                  The components are represented as PyMel objects.
        """
        objects_components_dict = {}

        for component in components_list:
            try:
                obj_name = component.node().name()
            except AttributeError:
                continue
            try:
                objects_components_dict[obj_name].append(component)
            except KeyError:
                objects_components_dict[obj_name] = [component]

        return objects_components_dict

    @classmethod
    def get_contigious_edges(cls, components):
        """Get a list containing sets of adjacent edges.

        Parameters:
            components (list): Polygon components to be filtered for adjacent edges.

        Returns:
            (list) adjacent edge sets.

        Example:
            get_contigious_edges(['obj.e[:2]']) #returns: [{'objShape.e[1]', 'objShape.e[0]', 'objShape.e[2]'}]
            get_contigious_edges(['obj.f[0]']) #returns: [{'objShape.e[24]', 'objShape.e[0]', 'objShape.e[25]', 'objShape.e[12]'}]
        """
        edges = cls.convert_component_type(components, "edge", flatten=1)

        sets = []
        for edge in edges:
            vertices = pm.polyListComponentConversion(edge, fromEdge=1, toVertex=1)
            connEdges = cls.convert_component_type(vertices, "edge", flatten=1)
            edge_set = set(
                [e for e in connEdges if e in edges]
            )  # restrict the connected edges to the original edge pool.
            sets.append(edge_set)

        result = []
        while len(sets) > 0:  # combine sets in 'sets' that share common elements.
            first, rest = sets[0], sets[1:]  # python 3: first, *rest = sets
            first = set(first)

            lf = -1
            while len(first) > lf:
                lf = len(first)

                rest2 = []
                for r in rest:
                    if len(first.intersection(set(r))) > 0:
                        first |= set(r)
                    else:
                        rest2.append(r)
                rest = rest2

            result.append(first)
            sets = rest

        return result

    @classmethod
    def get_contigious_islands(cls, faces):
        """Get a list containing sets of adjacent polygon faces grouped by islands.

        Parameters:
            faces (str/obj/list): The polygon faces to be filtered for adjacent.

        Returns:
            (list): of sets of adjacent faces.

        Example:
            get_contigious_islands('obj.f[21:26]') #returns: [{'objShape.f[22]', 'objShape.f[21]', 'objShape.f[23]'}, {'objShape.f[26]', 'objShape.f[24]', 'objShape.f[25]'}]
        """
        face_islands = []
        sets = []
        faces = pm.ls(faces, flatten=1)
        for face in faces:
            edges = pm.polyListComponentConversion(face, fromFace=1, toEdge=1)
            borderFaces = cls.convert_component_type(edges, "face", "obj", flatten=1)
            set_ = set([str(f) for f in borderFaces if f in faces])
            if set_:
                sets.append(set_)

        while len(sets) > 0:
            first, rest = sets[0], sets[1:]
            first = set(first)

            lf = -1
            while len(first) > lf:
                lf = len(first)
                rest2 = []
                for r in rest:
                    if len(first.intersection(set(r))) > 0:
                        first |= set(r)
                    else:
                        rest2.append(r)
                rest = rest2

            face_islands.append(first)
            sets = rest

        return face_islands

    @staticmethod
    def get_islands(obj, returned_type="str", flatten=False):
        """Get the group of components in each separate island of a combined mesh.

        Parameters:
            obj (str/obj/list): The object to get shells from.
            returned_type (bool): Return the shell faces as a list of type: 'str' (default), 'int', or 'obj'.
            flatten (bool): Flattens the returned list of objects so that each component is it's own element.

        Returns:
            (generator)

        Example:
            get_islands('combined_obj') #returns: [['combined_obj.f[0]', 'combined_obj.f[5]', ..etc, ['combined_obj.f[15]', ..etc]]
        """
        if isinstance(obj, str):
            obj = pm.ls(obj)[0]  # Convert input to a single PyMel object
        elif not isinstance(obj, pm.nt.Transform):
            raise ValueError(f"Expected a single mesh object, got {type(obj)}")

        # num_shells = pm.polyEvaluate(obj, shell=True)
        num_faces = pm.polyEvaluate(obj, face=True)
        unprocessed = set(range(num_faces))

        while unprocessed:
            # face_index
            index = next(iter(unprocessed))
            # shell faces
            faces = pm.polySelect(obj, extendToShell=index, noSelection=True)

            if returned_type == "str":
                yield ["{}.f[{}]".format(obj, index) for index in faces]

            elif returned_type == "int":
                yield [index for index in faces]

            elif returned_type == "obj":
                yield [pm.ls("{}.f[{}]".format(obj, index))[0] for index in faces]

            unprocessed.difference_update(faces)

    @classmethod
    def get_border_components(
        cls,
        x,
        component_type="",
        returned_type="str",
        component_border=False,
        flatten=False,
    ):
        """Get any object border components from given component(s) or a polygon object.
        A border is defined as a hole or detached edge.

        Parameters:
            x (str/obj/list): Component(s) (or a polygon object) to find any border components for.
            component_type (str): The desired returned component type. (valid: 'vertex','edge','face', '',
                    An empty string returns the same type as the first given component, or edges if an object is given)
            returned_type (str): The desired returned object type.
                    (valid: 'str'(default), 'obj'(shape object), 'transform'(as string), 'int'(valid only at sub-object level).
            component_border (bool): Get the components that border given components instead of the mesh border.
                    (valid: 'component', 'object'(default))
            flatten (bool): Flattens the returned list of objects so that each component is it's own element.

        Returns:
            (list) components that border an open edge.

        Example:
        get_border_components('pln', 'vtx') #returns: ['plnShape.vtx[0:4]', 'plnShape.vtx[7:8]', 'plnShape.vtx[11:15]'],
            get_border_components('pln') #returns: ['plnShape.e[0:2]', 'plnShape.e[4]', 'plnShape.e[6]', 'plnShape.e[8]', 'plnShape.e[13]', 'plnShape.e[15]', 'plnShape.e[20:23]'],
            get_border_components('pln.e[:]') #returns: ['plnShape.e[0:2]', 'plnShape.e[4]', 'plnShape.e[6]', 'plnShape.e[8]', 'plnShape.e[13]', 'plnShape.e[15]', 'plnShape.e[20:23]'],
            get_border_components(['pln.e[9]','pln.e[10]', 'pln.e[12]', 'pln.e[16]'], 'f', component_border=True) #returns: ['plnShape.f[1]', 'plnShape.f[3:5]', 'plnShape.f[7]'],
            get_border_components('pln.f[3:4]', 'vtx', component_border=True) #returns: ['plnShape.vtx[4:6]', 'plnShape.vtx[8:10]'],
        """
        if not x:
            raise ValueError(
                f"get_border_components expected object(s) or component(s), got {type(x)}"
            )

        origType = cls.get_component_type(x, "abv")
        if not origType:
            origType, x = "mesh", cls.get_components(x, "edges")
        origVerts = cls.convert_component_type(x, "vtx", flatten=True)
        origEdges = cls.convert_component_type(x, "edge", flatten=True)
        origFaces = cls.convert_component_type(x, "face", flatten=True)

        # If no component type is specified, return the same type of component as given. in the case of mesh object, edges will be returned.
        if not component_type:
            component_type = origType if not origType == "mesh" else "e"
        else:  # Get the correct component_type variable from possible args.
            component_type = cls.convert_alias(component_type)

        result = []
        if component_border:  # Get edges Qthat form the border of the given components.
            for edge in origEdges:
                attachedFaces = cls.convert_component_type(edge, "face", flatten=1)
                if component_type == "f":
                    for f in attachedFaces:
                        if origType == "f" and f in origFaces:
                            continue
                        result.append(f)
                    continue
                attachedEdges = cls.convert_component_type(
                    attachedFaces, "edge", flatten=1
                )
                for e in attachedEdges:
                    if origType == "e" and e in origEdges:
                        continue
                    attachedVerts = cls.convert_component_type(e, "vtx", flatten=1)
                    for v in attachedVerts:
                        if v in origVerts:
                            result.append(v)

        else:  # Get edges that form the border of the object.
            for edge in origEdges:
                attachedFaces = cls.convert_component_type(edge, "face", flatten=1)
                if len(attachedFaces) == 1:
                    result.append(edge)

        # Convert back to the original component type and flatten /un-flatten list.
        result = cls.convert_component_type(result, component_type)
        result = core_utils.CoreUtils.convert_array_type(
            result, returned_type=returned_type, flatten=flatten
        )
        return result

    @classmethod
    def get_closest_verts(cls, a, b, tolerance=1000):
        """Find the two closest vertices between the two sets of vertices.

        Parameters:
            a (str/list): The first set of vertices.
            b (str/list): The second set of vertices.
            tolerance (float) = Maximum search distance.

        Returns:
            (list): closest vertex pairs by order of distance (excluding those not meeting the tolerance). (<vertex from a>, <vertex from b>).

        Example:
            get_closest_verts('pln.vtx[:10]', 'pln.vtx[11:]', 6.667) #returns: [('plnShape.vtx[7]', 'plnShape.vtx[11]'), ('plnShape.vtx[8]', 'plnShape.vtx[12]'), ('plnShape.vtx[9]', 'plnShape.vtx[13]'), ('plnShape.vtx[10]', 'plnShape.vtx[11]'), ('plnShape.vtx[10]', 'plnShape.vtx[14]')]
        """
        from operator import itemgetter

        a = core_utils.CoreUtils.convert_array_type(
            a, returned_type="str", flatten=True
        )
        b = core_utils.CoreUtils.convert_array_type(
            b, returned_type="str", flatten=True
        )
        vertPairsAndDistance = {}
        for v1 in a:
            v1Pos = pm.pointPosition(v1, world=1)
            for v2 in b:
                v2Pos = pm.pointPosition(v2, world=1)
                distance = ptk.get_distance(v1Pos, v2Pos)
                if distance < tolerance:
                    vertPairsAndDistance[(v1, v2)] = distance

        sorted_ = sorted(vertPairsAndDistance.items(), key=itemgetter(1))
        vertPairs = [i[0] for i in sorted_]

        return vertPairs

    @classmethod
    @core_utils.CoreUtils.undo
    def get_closest_vertex(
        cls, vertices, obj, tolerance=0.0, freeze_transforms=False, returned_type="str"
    ):
        """Find the closest vertex of the given object for each vertex in the list of given vertices.

        Parameters:
            vertices (list): A set of vertices.
            obj (str/obj/list): The reference object in which to find the closest vertex for each vertex in the list of given vertices.
            tolerance (float) = Maximum search distance. Default is 0.0, which turns off the tolerance flag.
            freeze_transforms (bool): Reset the selected transform and all of its children down to the shape level.
            returned_type (str): The desired returned object type. This only affects the dict value (found vertex),
                            the key (orig vertex) is always a string. ex. {'planeShape.vtx[0]': 'objShape.vtx[3]'} vs. {'planeShape.vtx[0]': MeshVertex('objShape.vtx[3]')}
                            (valid: 'str'(default), 'obj'(shape object), 'transform'(as string), 'int'(valid only at sub-object level).
        Returns:
            (dict) closest vertex pairs {<vertex from a>:<vertex from b>}.

        Example:
            get_closest_vertex('plnShape.vtx[0]', 'cyl', returned_type='int') #returns: {'plnShape.vtx[0]': 3},
            get_closest_vertex('plnShape.vtx[0]', 'cyl') #returns: {'plnShape.vtx[0]': 'cylShape.vtx[3]'},
            get_closest_vertex('plnShape.vtx[2:3]', 'cyl') #returns: {'plnShape.vtx[2]': 'cylShape.vtx[2]', 'plnShape.vtx[3]': 'cylShape.vtx[1]'}
        """
        vertices = core_utils.CoreUtils.convert_array_type(
            vertices, returned_type="str", flatten=True
        )

        if freeze_transforms:
            pm.makeIdentity(obj, apply=True)

        # pm.listRelatives(obj, fullPath=False, shapes=True, noIntermediate=True)
        obj2Shape = pm.listRelatives(obj, children=1, shapes=1)[0]
        # create a closestPointOnMesh node.
        cpmNode = pm.ls(pm.createNode("closestPointOnMesh"))[0]
        # object's shape mesh output to the cpm node.
        pm.connectAttr(obj2Shape.outMesh, cpmNode.inMesh, force=1)

        closestVerts = {}
        for (
            v1
        ) in (
            vertices
        ):  # assure the list of vertices is a flattened list of stings. prevent unhashable type error when closestVerts[v1] = v2.  may not be needed with python versions 3.8+
            v1Pos = pm.pointPosition(v1, world=True)
            pm.setAttr(
                cpmNode.inPosition, v1Pos[0], v1Pos[1], v1Pos[2], type="double3"
            )  # set a compound attribute

            index = pm.getAttr(
                cpmNode.closestVertexIndex
            )  # vertex Index. | ie. result: [34]
            v2 = obj2Shape.vtx[index]

            v2Pos = pm.pointPosition(v2, world=True)
            distance = ptk.get_distance(v1Pos, v2Pos)

            v2_convertedType = core_utils.CoreUtils.convert_array_type(
                v2, returned_type=returned_type
            )[0]
            if not tolerance:
                closestVerts[v1] = v2_convertedType
            elif distance < tolerance:
                closestVerts[v1] = v2_convertedType

        pm.delete(cpmNode)

        return closestVerts

    @classmethod
    def get_edge_path(
        cls, components, path="edgeLoop", returned_type="str", flatten=False
    ):
        """Query the polySelect command for the components along different edge paths.
        Supports components from a single object.

        Parameters:
            components (str/obj/list): The components used for the query (dependant on the operation type).
            path (str): The desired return type. valid: 'edgeLoop': Select an edge loop starting at the given edge.
                    'edgeRing': Select an edge ring starting at the given edge.
                    'edgeRingPath', Given two edges that are on the same edge ring, this will select the shortest path between them on the ring.
                    'edgeLoopPath': Given two edges that are on the same edge loop, this will select the shortest path between them on the loop.
            returned_type (str): The desired returned object type.
                    (valid: 'str'(default), 'obj'(shape object), 'transform'(as string), 'int'(valid only at sub-object level).
            flatten (bool): Flattens the returned list of objects so that each component is it's own element.

        Returns:
            (list) The components comprising the path.

        Example:
            get_edge_path('sph.e[12]', 'edgeLoop') #returns: ['sphShape.e[12]', 'sphShape.e[17]', 'sphShape.e[16]', 'sphShape.e[15]', 'sphShape.e[14]', 'sphShape.e[13]']
            get_edge_path('sph.e[12]', 'edgeLoop', 'int') #returns: [12, 17, 16, 15, 14, 13]
            get_edge_path('sph.e[12]', 'edgeRing') #returns: ['sphShape.e[0]', 'sphShape.e[6]', 'sphShape.e[12]', 'sphShape.e[18]', 'sphShape.e[24]']
            get_edge_path(['sph.e[43]', 'sph.e[46]'], 'edgeRingPath') #returns: ['sphShape.e[43]', 'sphShape.e[42]', 'sphShape.e[47]', 'sphShape.e[46]']
            get_edge_path(['sph.e[54]', 'sph.e[60]'], 'edgeLoopPath') #returns: ['sphShape.e[60]', 'sphShape.e[48]', 'sphShape.e[42]', 'sphShape.e[36]', 'sphShape.e[30]', 'sphShape.e[54]']
        """
        obj, *other = pm.ls(components, objectsOnly=1)
        cnums = cls.convert_component_type(
            components, "edge", returned_type="int", flatten=True
        )

        if len(cnums) < 2 and path in ("edgeRingPath", "edgeLoopPath"):
            print(
                f'File "{__file__}" in get_edge_path\n# Error: Operation requires at least two components. #\n Edges given: {cnums}',
            )
            return []

        if path == "edgeRing":
            edgesLong = pm.polySelect(obj, q=True, edgeRing=cnums)  # (e..)

        elif path == "edgeRingPath":
            edgesLong = pm.polySelect(
                obj, q=True, edgeRingPath=(cnums[0], cnums[1])
            )  # (e, e)
            if not edgesLong:
                print(
                    f'File "{__file__}" in get_edge_path\n# Error: get_edge_path: Operation requires two edges that are on the same edge ring.\n\tEdges given: {cnums[0]}, {cnums[1]}',
                )
                return []

        elif path == "edgeLoopPath":
            edgesLong = pm.polySelect(
                obj, q=True, edgeLoopPath=(cnums[0], cnums[1])
            )  # (e, e)
            if not edgesLong:
                print(
                    f'File "{__file__}" in get_edge_path\n# Error: get_edge_path: Operation requires two edges that are on the same edge loop.\n\tEdges given: {cnums[0]}, {cnums[1]}',
                )
                return []
        else:  # EdgeLoop
            edgesLong = pm.polySelect(obj, q=True, edgeLoop=cnums)  # (e..)

        objName = obj.name()
        result = ptk.remove_duplicates(
            ["{}.e[{}]".format(objName, e) for e in edgesLong]
        )
        return core_utils.CoreUtils.convert_array_type(
            result, returned_type=returned_type, flatten=flatten
        )

    @classmethod
    def get_shortest_path(cls, components, returned_type="str", flatten=False):
        """Get the shortest path between two components.

        Parameters:
            components (obj): A Pair of vertices or edges.
            returned_type (str): The desired returned object type.
                    valid: 'str'(default), 'obj', 'int'(valid only at sub-object level)
            flatten (bool): Flattens the returned list of objects so that each component is it's own element.

        Returns:
            (list) the components that comprise the path as strings.

        Example:
            get_edge_path('sph.e[12]', 'edgeLoop') #returns: ['sphShape.e[12]', 'sphShape.e[17]', 'sphShape.e[16]', 'sphShape.e[15]', 'sphShape.e[14]', 'sphShape.e[13]']
            get_edge_path('sph.e[12]', 'edgeLoop', 'int') #returns: [12, 17, 16, 15, 14, 13]
            get_edge_path('sph.e[12]', 'edgeRing') #returns: ['sphShape.e[0]', 'sphShape.e[6]', 'sphShape.e[12]', 'sphShape.e[18]', 'sphShape.e[24]']
            get_edge_path(['sph.e[43]', 'sph.e[46]'], 'edgeRingPath') #returns: ['sphShape.e[43]', 'sphShape.e[42]', 'sphShape.e[47]', 'sphShape.e[46]']
            get_edge_path(['sph.e[54]', 'sph.e[60]'], 'edgeLoopPath') #returns: ['sphShape.e[60]', 'sphShape.e[48]', 'sphShape.e[42]', 'sphShape.e[36]', 'sphShape.e[30]', 'sphShape.e[54]']
        """
        obj = pm.ls(components, objectsOnly=1)[0]
        ctype = cls.get_component_type(components)
        try:
            A, B = components = cls.convert_component_type(components, ctype)[:2]
        except ValueError as e:
            print(
                f'File "{__file__}" in get_shortest_path\n# Error: Operation requires exactly two components.\n\t{e}',
            )
            return []

        returnAsVerts = False
        if ctype == "vtx":
            edgesA = cls.convert_component_type(A, "e", flatten=1)
            vertsA = cls.convert_component_type(edgesA, "vtx", flatten=1)
            closestA = cls.get_closest_verts(B, [i for i in vertsA if not i == A])[0]
            edgeA = [
                e
                for e in edgesA
                if closestA[1] in cls.convert_component_type(e, "vtx", flatten=1)
            ]

            edgeB = cls.convert_component_type(B, "e", flatten=1)
            vertsB = cls.convert_component_type(edgeB, "vtx", flatten=1)
            closestB = cls.get_closest_verts(A, [i for i in vertsB if not i == B])[0]
            edgeB = [
                e
                for e in edgeB
                if closestB[1] in cls.convert_component_type(e, "vtx", flatten=1)
            ]

            components = (edgeA, edgeB)
            ctype = "e"
            returnAsVerts = True

        compNums = cls.convert_component_type(
            components, ctype, returned_type="int", flatten=True
        )

        kwargs = {
            "shortestFacePath"
            if ctype == "f"
            else "shortestEdgePathUV"
            if ctype == "uv"
            else "shortestEdgePath": compNums
        }
        compLong = set(pm.polySelect(obj, q=True, **kwargs) + compNums)

        result = cls.convert_int_to_component(
            obj, compLong, ctype, returned_type=returned_type, flatten=flatten
        )

        if returnAsVerts:
            result = cls.convert_component_type(result, "vtx", flatten=flatten)

        return result

    @staticmethod
    def get_normal(face):
        """Get the normal of a face in world space.

        Parameters:
            face (pymel.core.nodetypes.MeshFace): The face to get the normal of.

        Returns:
            om.MVector: The normal of the face in world space.

        Raises:
            TypeError: If the input is not a MeshFace.
        """
        import maya.OpenMaya as om

        if not isinstance(face, pm.general.MeshFace):
            raise TypeError(f"Input must be a MeshFace, got {type(face)}.")

        # Create an MSelectionList and add the face to it
        sel_list = om.MSelectionList()
        sel_list.add(face.name())

        # Create an MDagPath and an MObject for the face
        dag_path = om.MDagPath()
        component = om.MObject()
        sel_list.getDagPath(0, dag_path, component)

        # Create an MFnMesh for the mesh the face belongs to
        mesh_fn = om.MFnMesh(dag_path)

        # Get the index of the face
        face_index = int(face.name().split("[")[-1].split("]")[0])

        # Get the normal of the face
        normal = om.MVector()
        mesh_fn.getPolygonNormal(face_index, normal, om.MSpace.kWorld)

        return normal

    @staticmethod
    def get_normal_vector(x):
        """Get the normal vectors of the given polygon object(s) or its components.

        Parameters:
            x (str/obj/list): A polygon mesh or its components. Accepts a string representation, a PyNode object,
                    or a list of either, representing one or more polygon meshes or their faces.
        Returns:
            dict: A dictionary where each key-value pair corresponds to a face of the polygon object(s).
                  The key is the face's ID, and the value is a list representing the face's normal vector in the format [x, y, z].

        This function extracts the normal vectors for each face of the provided polygon object(s).
        If components (faces) are provided directly, the function will only calculate and return the normals for those faces.
        """
        obj = pm.ls(x)
        normals = pm.polyInfo(obj, faceNormals=1)

        regex = "[A-Z]*_[A-Z]* *[0-9]*: "

        dct = {}
        for n in normals:
            lst = list(
                s.replace(regex, "") for s in n.split() if s
            )  # ['FACE_NORMAL', '150:', '0.935741', '0.110496', '0.334931\n']

            key = int(lst[1].strip(":"))  # int face number as key ie. 150
            value = list(
                float(i) for i in lst[-3:]
            )  # vector list as value. ie. [[0.935741, 0.110496, 0.334931]]
            dct[key] = value

        return dct

    @classmethod
    def get_normal_angle(cls, edge):
        """Get the angle between the normals of the two faces connected by an edge.

        Parameters:
            edge (MeshEdge): The edge to get the normal angle of.

        Returns:
            float: The angle between the normals of the two faces connected by the edge, in degrees.

        Raises:
            TypeError: If the input edge is not a MeshEdge.
        """
        import math

        if not isinstance(edge, pm.general.MeshEdge):
            raise TypeError(f"Input must be a MeshEdge, got {type(edge)}.")

        # Get the faces connected by the edge
        connected_faces = list(edge.connectedFaces())
        if len(connected_faces) != 2:
            return 0

        # Get the normals of the faces
        normal1 = cls.get_normal(connected_faces[0])
        normal2 = cls.get_normal(connected_faces[1])
        # Calculate the angle between the normals
        angle = normal1.angle(normal2)
        # Convert the angle from radians to degrees
        angle = math.degrees(angle)

        return angle

    @classmethod
    @core_utils.CoreUtils.undo
    def average_normals(cls, objects, by_uv_shell=False):
        """Average the normals of the given objects.

        Parameters:
            objects (str/obj/list): The mesh or mesh faces to averge.
            by_uv_shell (bool): Average each UV shell individually.
        """
        # Map components to their respective objects
        components_dict = cls.map_components_to_objects(objects)

        # Loop through each object and its corresponding components
        for obj, components in components_dict.items():
            if by_uv_shell:
                uv_shell_sets = cls.get_uv_shell_sets(components)
                for uv_set in uv_shell_sets:
                    pm.polySoftEdge(uv_set, a=180)
            else:
                if components:  # if faces/components are selected
                    pm.polySoftEdge(components, a=180)
                else:  # if objects are selected
                    pm.polySoftEdge(obj, a=180)

    @classmethod
    def get_edges_by_normal_angle(cls, objects, low_angle=0, high_angle=180):
        """Get a list of edges having normals between the given high and low angles.

        Parameters:
            objects (str/list)(obj): The object(s) to get edges of.
            low_angle (int): Normal angle low range.
            high_angle (int): Normal angle high range.

        Returns:
            list: Polygon edges that have normals within the specified angle range.

        Raises:
            TypeError: If the input objects are not Mesh or MeshEdge.
        """
        # Assure objects is a list
        objects = pm.ls(objects)

        edges = []
        # Iterate over the objects
        for obj in objects:
            # If the object is a transform, get its shape
            if isinstance(obj, pm.nt.Transform):
                obj = obj.getShape()

            # If the object is a mesh, get all its edges
            if isinstance(obj, pm.nt.Mesh):
                edges.extend(obj.edges)

            # If the object is an edge, add it to the list
            elif isinstance(obj, pm.general.MeshEdge):
                edges.append(obj)

            else:
                raise TypeError(f"Input must be a Mesh or MeshEdge, got {type(obj)}.")

        # Filter the edges based on their normal angle
        edges = [
            edge
            for edge in edges
            if low_angle <= cls.get_normal_angle(edge) <= high_angle
        ]

        return edges

    @classmethod
    @core_utils.CoreUtils.undo
    def set_edge_hardness(
        cls, x, angle_threshold, upper_hardness=None, lower_hardness=None
    ):
        """Sets the hardness (softness) of edges in the provided objects based on their normal angles.
        The function recursively processes lists, tuples, and sets of objects, applying the hardness settings
        to each item individually. It also supports PyMel Transform, Mesh, and MeshEdge objects.
        If an edge's normal angle is greater than or equal to the given threshold, the edge is considered
        for upper hardness application. If an edge's normal angle is less than the threshold, the edge is
        considered for lower hardness application.

        Parameters:
            cls (class): The class that the method is part of.
            x (str, pm.nt.Transform, pm.nt.Mesh, pm.general.MeshEdge, list/tuple/set of these types):
                The objects whose edge hardness is to be set. For string input, it should be in the format 'object.e[start:end]'.
            angle_threshold (float): The threshold of the normal angle in degrees to determine hardness.
            upper_hardness (float, optional): The hardness to apply to edges with a normal angle greater
                than or equal to the threshold. If None, these edges are not processed. Value should be between 0 and 180.
            lower_hardness (float, optional): The hardness to apply to edges with a normal angle less
                than the threshold. If None, these edges are not processed. Value should be between 0 and 180.

        Returns:
            None: This function doesn't return anything; it modifies the provided objects in-place.

        Raises:
            TypeError: If the 'x' argument is not of the correct type.
        """
        if isinstance(x, (list, tuple, set)):
            for item in x:
                cls.set_edge_hardness(
                    item, angle_threshold, upper_hardness, lower_hardness
                )
            return

        if isinstance(x, pm.nt.Transform):
            is_group = node_utils.NodeUtils.is_group(x)
            if is_group:
                grp_children = node_utils.NodeUtils.get_unique_children(x)
                cls.set_edge_hardness(
                    grp_children, angle_threshold, upper_hardness, lower_hardness
                )
                return

            shape = x.getShape()
            if not isinstance(shape, pm.nt.Mesh):
                raise TypeError(
                    f"Transform node {x} has a shape of unsupported type {type(shape)}"
                )

            x = f"{shape.name()}.e[0:{len(shape.edges)-1}]"

        elif isinstance(x, pm.nt.Mesh):
            x = f"{x.name()}.e[0:{len(x.edges())-1}]"

        elif isinstance(x, pm.general.MeshEdge):
            x = x.name()

        else:
            print(f"Unsupported type {type(x)}")
            return

        if isinstance(x, str):
            x = [x]

        upper_hardness_edges = []
        lower_hardness_edges = []

        for obj_name in x:
            mesh_obj = pm.PyNode(obj_name.split(".")[0])
            edge_indices = obj_name.split("[")[-1].split("]")[0].split(":")
            edge_start = int(edge_indices[0])
            edge_end = int(edge_indices[1]) if len(edge_indices) > 1 else edge_start
            edge_indices = list(range(edge_start, edge_end + 1))

            for edge_index in edge_indices:
                edge = mesh_obj.edges[edge_index]
                edge_angle = cls.get_normal_angle(edge)
                if upper_hardness is not None and edge_angle >= angle_threshold:
                    upper_hardness_edges.append(edge)
                elif lower_hardness is not None and edge_angle < angle_threshold:
                    lower_hardness_edges.append(edge)

        if upper_hardness_edges:
            pm.polySoftEdge(upper_hardness_edges, a=upper_hardness, ch=True)

        if lower_hardness_edges:
            pm.polySoftEdge(lower_hardness_edges, a=lower_hardness, ch=True)

    @classmethod
    def get_faces_with_similar_normals(
        cls,
        faces,
        transforms=[],
        similar_faces=[],
        range_x=0.1,
        range_y=0.1,
        range_z=0.1,
        returned_type="str",
    ):
        """Filter for faces with normals that fall within an X,Y,Z tolerance.

        Parameters:
            faces (list): ['polygon faces'] - faces to find similar normals for.
            similar_faces (list): optional ability to add faces from previous calls to the return value.
            transforms (list): [<shape nodes>] - objects to check faces on. If none are given the objects containing the given faces will be used.
            range_x = float - x axis tolerance
            range_y = float - y axis tolerance
            range_z = float - z axis tolerance
            returned_type (str): The desired returned object type.
                        valid: 'str'(default), 'obj'(shape object), 'transform'(as string), 'int'(valid only at sub-object level).
        Returns:
            (list) faces that fall within the given normal range.

        Example:
            get_faces_with_similar_normals(selectedFaces, range_x=0.5, range_y=0.5, range_z=0.5)
        """
        # Work on a copy of the argument so that removal of elements doesn't effect the passed in list.
        faces = pm.ls(faces, flatten=1)
        for face in faces:
            normals = cls.get_normal_vector(face)

            for k, v in normals.items():
                sX, sY, sZ = v

                if not transforms:
                    transforms = pm.ls(face, objectsOnly=True)

                for node in transforms:
                    for f in cls.get_components(
                        node, "faces", returned_type=returned_type, flatten=1
                    ):
                        n = cls.get_normal_vector(f)
                        for k, v in n.items():
                            nX, nY, nZ = v

                            if (
                                sX <= nX + range_x
                                and sX >= nX - range_x
                                and sY <= nY + range_y
                                and sY >= nY - range_y
                                and sZ <= nZ + range_z
                                and sZ >= nZ - range_z
                            ):
                                similar_faces.append(f)
                                if (
                                    f in faces
                                ):  # If the face is in the loop que, remove it, as has already been evaluated.
                                    faces.remove(f)

        return similar_faces

    @staticmethod
    @core_utils.CoreUtils.undo
    def transfer_normals(source, target):
        """Transfer normal information from one object to another.

        Parameters:
            source (str/obj/list): The transform node to copy normals from.
            target (str/obj/list): The transform node(s) to copy normals to.
        """
        s, *other = pm.ls(source)
        # store source transforms
        sourcePos = pm.xform(s, q=1, t=1, ws=1)
        sourceRot = pm.xform(s, q=1, ro=1, ws=1)
        sourceScale = pm.xform(s, q=1, s=1, ws=1)

        for t in pm.ls(target):
            # store target transforms
            targetPos = pm.xform(t, q=1, t=1, ws=1)
            targetRot = pm.xform(t, q=1, ro=1, ws=1)
            targetScale = pm.xform(t, q=1, s=1, ws=1)

            # move target to source position
            pm.xform(t, t=sourcePos, ws=1)
            pm.xform(t, ro=sourceRot, ws=1)
            pm.xform(t, s=sourceScale, ws=1)

            # copy normals
            pm.polyNormalPerVertex(t, ufn=0)
            pm.transferAttributes(s, t, pos=0, nml=1, uvs=0, col=0, spa=0, sm=3, clb=1)
            pm.delete(t, ch=1)

            # restore t position
            pm.xform(t, t=targetPos, ws=1)
            pm.xform(t, ro=targetRot, ws=1)
            pm.xform(t, s=targetScale, ws=1)

    @classmethod
    def filter_components_by_connection_count(
        cls, components, num_of_connected=(0, 2), connected_type="", returned_type="str"
    ):
        """Get a list of components filtered by the number of their connected components.

        Parameters:
            components (str/list)(obj): The components to filter.
            num_of_connected (int)(tuple): The number of connected components. Can be given as a range. (Default: (0,2))
            connected_type (str)(int): The desired component mask. (valid: 'vtx','vertex','vertices','Polygon Vertex',31,0x0001(vertices), 'e','edge','edges','Polygon Edge',32,0x8000(edges), 'f','face','faces','Polygon Face',34,0x0008(faces), 'uv','texture','texture coordinates','Polygon UV',35,0x0010(texture coordiantes).
            returned_type (str): The desired returned object type.
                        valid: 'str'(default), 'obj', 'int'(valid only at sub-object level)
        Returns:
            (list) flattened list.

        Example:
            faces = filter_components_by_connection_count('sph.f[:]', 4, 'e') #returns faces with four connected edges (four sided faces).
            verts = filter_components_by_connection_count('pln.vtx[:]', (0,2), 'e') #returns vertices with up to two connected edges.
        """
        if isinstance(num_of_connected, (tuple, list)) and len(num_of_connected) == 2:
            lowRange, highRange = num_of_connected
        elif isinstance(num_of_connected, int):
            lowRange = highRange = num_of_connected
        else:
            raise TypeError(
                f"num_of_connected expected an int or two int tuple, got {type(num_of_connected)}"
            )

        typ = cls.get_component_type(components)
        if connected_type:
            ctype = cls.convert_alias(connected_type)
        else:
            ctype = typ

        result = []
        for c in pm.ls(components, flatten=True):
            attached = cls.convert_component_type(c, ctype, flatten=True)
            n = len(attached)
            if n >= lowRange and n <= highRange:
                result.append(c)

        result = core_utils.CoreUtils.convert_array_type(
            result, returned_type=returned_type
        )
        return result

    @classmethod
    def get_vertex_normal(cls, vertex, angle_weighted=False):
        """Return the normal at the given vertex. The returned normal is a single
        per-vertex normal, so unshared normals at a vertex will be averaged.

        Parameters:
            vertex (str/obj/list): A polygon vertex.
            angle_weighted (bool): Weight by the angle subtended by the face at the vertex.
                    If angle_weighted is set to false, a simple average of surround face normals is returned.
                    The simple average evaluation is significantly faster than the angle-weighted average.
        Returns:
            (MVector)
        """
        import maya.api.OpenMaya as om

        mesh = pm.ls(vertex, objectsOnly=True)[0].name()
        selectionList = om.MSelectionList()  # empty selection list.
        selectionList.add(mesh)

        dagPath = selectionList.getDagPath(0)  # create empty dag path object.
        mesh = om.MFnMesh(dagPath)  # get mesh.

        vtxID = core_utils.CoreUtils.convert_array_type(vertex, "int")[0]
        # get vertex normal and use om.MSpace.kObject for object space.
        return mesh.getVertexNormal(vtxID, angle_weighted, space=om.MSpace.kWorld)

    @staticmethod
    def get_vector_from_components(components):
        """Get a vector representing the averaged and normalized vertex-face normals.

        Parameters:
            components (list): A list of component to get normals of.

        Returns:
            (tuple) vector ie. (-4.5296159711938344e-08, 1.0, 1.6846732009412335e-08)
        """
        vertices = pm.polyListComponentConversion(components, toVertex=1)

        norm = pm.polyNormalPerVertex(
            vertices, query=True, xyz=True
        )  # return all of the normals associated with the vert.
        normal_vector = (
            sum(norm[0::3]) / len(norm[0::3]),
            sum(norm[1::3]) / len(norm[1::3]),
            sum(norm[2::3]) / len(norm[2::3]),
        )  # averaging of all x,y,z points.

        return normal_vector

    @staticmethod
    def orient_shells(objects):
        """Rotate UV shells to run parallel with the most adjacent U or V axis of their bounding box.

        Parameters:
            objects (str/obj/list): Polygon mesh objects and/or components.
        """
        for obj in pm.ls(objects, objectsOnly=1):
            # filter components for only this object.
            obj_compts = [i for i in objects if obj in pm.ls(i, objectsOnly=1)]
            pm.polyLayoutUV(
                obj_compts,
                flipReversed=0,
                layout=0,
                layoutMethod=1,
                percentageSpace=0.2,
                rotateForBestFit=3,
                scale=0,
                separate=0,
            )

    @classmethod
    def move_to_uv_space(cls, objects, u, v, relative=True):
        """Move objects to the given u and v coordinates.

        Parameters:
            objects (str/obj/list): The object(s) to move.
            u (int): u coordinate.
            v (int): v coordinate.
            relative (bool): Move relative or absolute.
        """
        # Convert the objects to UVs
        uvs = pm.polyListComponentConversion(objects, fromFace=True, toUV=True)
        uvs = pm.ls(uvs, flatten=True)

        # Move the UVs to the given u and v coordinates
        pm.polyEditUV(uvs, u=u, v=v, relative=relative)

    @classmethod
    def get_uv_shell_sets(cls, objects=None, returned_type="shells"):
        """Get UV shells and their corresponding sets of faces.

        Parameters:
            objects (obj/list): Polygon object(s) or Polygon face(s).
            returned_type (str): The desired returned type. valid values are: 'shells', 'IDs'. If None is given, the full dict will be returned.

        Returns:
            (list)(dict) dependant on the given returned_type arg. ex. {0L:[[MeshFace(u'pShape.f[0]'), MeshFace(u'pShape.f[1]')], 1L:[[MeshFace(u'pShape.f[2]'), MeshFace(u'pShape.f[3]')]}
        """
        faces = cls.get_components(objects, "faces", flatten=1)

        shells = {}
        for face in faces:
            shell_Id = pm.polyEvaluate(face, uvShellIds=True)

            try:
                shells[shell_Id[0]].append(face)
            except KeyError:
                try:
                    shells[shell_Id[0]] = [face]
                except IndexError:
                    pass

        if returned_type == "shells":
            shells = list(shells.values())
        elif returned_type == "IDs":
            shells = shells.keys()

        return shells

    @staticmethod
    def get_uv_shell_border_edges(objects):
        """Get the edges that make up any UV islands of the given objects.

        Parameters:
            objects (str/obj/list): Polygon objects, mesh UVs, or Edges.

        Returns:
            (list): UV border edges.
        """
        uv_border_edges = []
        for obj in pm.ls(objects):
            # If the obj is a mesh object, get its shape
            if isinstance(obj, pm.nt.Transform):
                obj = obj.getShape()

            # If the obj is a mesh shape, get its UV borders
            if isinstance(obj, pm.nt.Mesh):
                # Get the connected edges to the selected UVs
                connected_edges = pm.polyListComponentConversion(
                    obj, fromUV=True, toEdge=True
                )
                connected_edges = pm.ls(connected_edges, flatten=True)
            elif isinstance(obj, pm.general.MeshEdge):
                # If the object is already an edge, no conversion is necessary
                connected_edges = pm.ls(obj, flatten=True)
            elif isinstance(obj, pm.general.MeshUV):
                # If the object is a UV, convert it to its connected edges
                connected_edges = pm.polyListComponentConversion(
                    obj, fromUV=True, toEdge=True
                )
                connected_edges = pm.ls(connected_edges, flatten=True)
            else:
                raise ValueError(f"Unsupported object type: {type(obj)}")

            for edge in connected_edges:
                edge_uvs = pm.ls(
                    pm.polyListComponentConversion(edge, tuv=True), fl=True
                )
                edge_faces = pm.ls(
                    pm.polyListComponentConversion(edge, tf=True), fl=True
                )
                if (
                    len(edge_uvs) > 2 or len(edge_faces) < 2
                ):  # If an edge has more than two uvs or less than 2 faces, it's a uv border edge.
                    uv_border_edges.append(edge)

        return uv_border_edges

    @staticmethod
    def transfer_uvs(source, target):
        """ """
        import maya.api.OpenMaya as om2

        # Create a MSelectionList
        sList = om2.MSelectionList()

        # Add objects to the list (accepts both PyNodes and strings)
        if isinstance(source, str):
            sList.add(source)
        else:
            sList.add(source.name())
        if isinstance(target, str):
            sList.add(target)
        else:
            sList.add(target.name())

        # Get MObjects of source and target
        source_obj = sList.getDagPath(0).extendToShape()  # extend to shape node
        target_obj = sList.getDagPath(1).extendToShape()  # extend to shape node

        # Get the function set for source and target
        sFnMesh = om2.MFnMesh(source_obj)
        tFnMesh = om2.MFnMesh(target_obj)

        # Get uv count for source
        uArray, vArray = sFnMesh.getUVs()

        # Clear target UVs and set them with source UVs
        tFnMesh.clearUVs()
        tFnMesh.setUVs(uArray, vArray)

        # Get the number of polygons in source mesh
        poly_count = sFnMesh.numPolygons

        # Assign UVs to faces
        for i in range(poly_count):
            count = sFnMesh.polygonVertexCount(i)
            for j in range(count):
                uv_id = sFnMesh.getPolygonUVid(i, j)
                tFnMesh.assignUV(i, j, uv_id)


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    pass

# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------



# !/usr/bin/python
# coding=utf-8
try:
    import pymel.core as pm
except ImportError as error:
    print(__file__, error)


class Preview:
    def __init__(self, *args, **kwargs):
        """Provides an interactive layer for previewing and finalizing operations in a 3D editing environment.

        This class enables real-time previews of operations by linking UI elements to backend functionality.
        It efficiently manages the state and execution of operations, maintaining a clean undo stack and enabling
        rollback of changes during the preview process.

        Parameters:
            operation_instance: Instance implementing a perform_operation method.
            preview_checkbox: QCheckBox instance to toggle the preview.
            create_button: QPushButton instance to finalize the changes.
            finalize_func: Optional callable to finalize changes.
            message_func: Optional callable for messaging, default is print.
            enable_on_show: Boolean, if True enables the preview when the window shows.
            disable_on_hide: Boolean, if True disables the preview when the window hides.

        Usage Example:
            ```python
            class BevelEdgesSlots:
                def __init__(self):
                    self.sb = self.switchboard()
                    self.ui = self.sb.bevel_edges
                    self.preview = Preview(
                        self,
                        self.ui.chk000,
                        self.ui.b000,
                        message_func=self.sb.message_box
                    )
                    self.sb.connect_multi(self.ui, "s000-1", "valueChanged", self.preview.refresh)

                def perform_operation(self, objects):
                    width = self.ui.s000.value()
                    segments = self.ui.s001.value()
                    BevelEdges.bevel_edges(objects, width, segments)

            # Instantiate BevelEdgesSlots.
            # Now toggling the UI checkbox will enable/disable the preview,
            # and clicking the UI button will apply the beveled edges.
            ```
        """
        self.init_preview(*args, **kwargs)

    def init_preview(
        self,
        operation_instance,
        preview_checkbox,
        create_button,
        finalize_func=None,
        message_func=print,
        enable_on_show=False,
        disable_on_hide=True,
    ):
        """Initialize the state for preview operations."""
        self.operated_objects = set()
        self.operation_performed = False
        self.needs_undo = False
        self.prevState = None
        self.operation_instance = operation_instance
        self.operation_instance.operated_objects = self.operated_objects
        self.preview_checkbox = preview_checkbox
        self.create_button = create_button
        self.finalize_func = finalize_func
        self.message_func = message_func
        self.preview_checkbox.clicked.connect(self.toggle)
        self.create_button.clicked.connect(self.finalize_changes)
        self.window = self.create_button.window()

        self.init_show_hide_behavior(enable_on_show, disable_on_hide)

    def init_show_hide_behavior(self, enable_on_show, disable_on_hide):
        self.enable_on_show = enable_on_show
        self.disable_on_hide = disable_on_hide

        if hasattr(self.window, "on_show"):
            self.window.on_show.connect(self.conditionally_enable)
        if hasattr(self.window, "on_hide"):
            self.window.on_hide.connect(self.conditionally_disable)

    def conditionally_enable(self):
        if self.enable_on_show:
            self.enable()

    def conditionally_disable(self):
        if self.disable_on_hide:
            self.disable()

    def toggle(self, state):
        """Toggles the preview on or off.

        Parameters:
            state: Boolean state to set.
        """
        if state:
            self.enable()
        else:
            self.disable()

    def enable(self):
        """Enables the preview and sets up the initial state."""
        self.prevState = pm.undoInfo(q=True, state=True)
        pm.undoInfo(state=True)
        pm.undoInfo(openChunk=True, chunkName="PreviewChunk")

        try:
            selected_objs = pm.selected()
            if selected_objs:
                self.operated_objects.update(selected_objs)

                self.needs_undo = False  # Set to False when enabling for the first time

                self.preview_checkbox.blockSignals(True)
                self.preview_checkbox.setChecked(True)
                self.preview_checkbox.blockSignals(False)

                self.create_button.setEnabled(True)
                self.refresh()
                self.operation_performed = True
            else:
                self.message_func("No objects selected.")
                self.disable()
        except Exception as e:
            print(f"Exception in enable: {e}")

    def disable(self):
        """Disables the preview and reverts to the initial state."""
        self.undo_if_needed()
        pm.undoInfo(closeChunk=True)

        if self.prevState is not None:
            pm.undoInfo(state=self.prevState)

        self.operated_objects.clear()
        self.preview_checkbox.setChecked(False)
        self.create_button.setEnabled(False)

    def undo_if_needed(self):
        """Executes undo operation if required."""
        if self.needs_undo:
            pm.undoInfo(closeChunk=True)
            try:
                pm.undo()
            except RuntimeError:
                pass
            finally:
                pm.undoInfo(openChunk=True, chunkName="PreviewChunk")

            self.needs_undo = False

    def refresh(self, *args):
        """Refreshes the preview to reflect any changes."""
        if not self.preview_checkbox.isChecked():
            return
        self.undo_if_needed()
        pm.undoInfo(openChunk=True, chunkName="PreviewChunk")
        try:
            self.operation_instance.perform_operation(self.operated_objects)
        except Exception as e:
            print(f"Exception during operation: {e}")
        finally:
            pm.undoInfo(closeChunk=True)
        self.needs_undo = True  # Set to True once the operation has been performed

    def finalize_changes(self):
        """Finalizes the preview changes and calls the finalize_func if provided."""
        self.needs_undo = False
        self.disable()
        if self.finalize_func:
            self.finalize_func()


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    pass

# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------



# !/usr/bin/python
# coding=utf-8
import unittest
import pymel.core as pm
import mayatk as mtk


class EditUtilsTest(unittest.TestCase):
    @classmethod
    def setUp(self):
        """Set up test scene once for all tests."""
        pm.mel.file(new=True, force=True)
        self.cube1 = pm.polyCube(
            width=5,
            height=5,
            depth=5,
            subdivisionsX=1,
            subdivisionsY=1,
            subdivisionsZ=1,
            name="cube1",
        )[0]
        self.cube2 = pm.polyCube(
            width=2,
            height=4,
            depth=8,
            subdivisionsX=3,
            subdivisionsY=3,
            subdivisionsZ=3,
            name="cube2",
        )[0]
        self.cyl = pm.polyCylinder(
            radius=5,
            height=10,
            subdivisionsX=12,
            subdivisionsY=1,
            subdivisionsZ=1,
            name="cyl",
        )[0]

    def test_rename(self):
        """Test renaming objects."""
        mtk.rename("cube1", "newName")
        self.assertTrue(pm.objExists("newName"))
        mtk.rename("newName", "cube1")
        self.assertTrue(pm.objExists("cube1"))

    def test_set_case(self):
        """Test setting the case of object names."""
        mtk.set_case("cube1", "upper")
        self.assertTrue(pm.objExists("CUBE1"))

    def test_append_location_based_suffix(self):
        """Test appending location-based suffixes to object names."""
        # Assuming the function name was incorrect and correcting it to match the function call.
        mtk.append_location_based_suffix(["cube1", "cube2"])
        self.assertTrue(pm.objExists("cube_0"))
        self.assertTrue(pm.objExists("cube_1"))

    def test_snap_closest_verts(self):
        """Test snapping closest vertices."""
        mtk.snap_closest_verts("cube1", "cube2")
        # Assuming the function modifies the geometry, so no specific return value to check.

    def test_merge_vertices(self):
        """Test merging vertices."""
        mtk.merge_vertices("cube1")
        # Assuming the function modifies the geometry, so no specific return value to check.

    def test_delete_along_axis(self):
        """Test deleting geometry along a specified axis."""
        mtk.delete_along_axis("cube1")
        # Assuming the function modifies the geometry, so no specific return value to check.

    def test_get_all_faces_on_axis(self):
        """Test getting all faces on a specified axis."""
        result = mtk.get_all_faces_on_axis("cube1")
        self.assertEqual(
            result,
            ["cube1.f[0]", "cube1.f[1]", "cube1.f[2]", "cube1.f[3]", "cube1.f[5]"],
        )

    def test_clean_geometry(self):
        """Test cleaning geometry."""
        mtk.clean_geometry("cyl")
        # Assuming the function modifies the geometry, so no specific return value to check.

    def test_get_overlapping_dup_objects(self):
        """Test getting overlapping duplicate objects."""
        result = mtk.get_overlapping_duplicates(["cyl", "cube1", "cube2"])
        self.assertEqual(result, set())

    def test_find_non_manifold_vertex(self):
        """Test finding non-manifold vertices."""
        result = mtk.find_non_manifold_vertex("cyl")
        self.assertEqual(result, set())

    def test_split_non_manifold_vertex(self):
        """Test splitting non-manifold vertices."""
        mtk.split_non_manifold_vertex("cyl")
        # Assuming the function modifies the geometry, so no specific return value to check.

    def test_get_ngons(self):
        """Test getting n-gons."""
        result = mtk.get_ngons("cyl")
        self.assertEqual(result, [])

    def test_get_overlapping_vertices(self):
        """Test getting overlapping vertices."""
        result = mtk.get_overlapping_vertices("cyl")
        self.assertEqual(result, [])

    def test_get_overlapping_faces(self):
        """Test getting overlapping faces."""
        result = mtk.get_overlapping_faces("cyl")
        self.assertEqual(result, [])
        result = mtk.get_overlapping_faces("cyl.f[:]")
        self.assertEqual(result, [])

    def test_get_similar_mesh(self):
        """Test finding similar mesh."""
        result = mtk.get_similar_mesh("cyl")
        self.assertEqual(result, [])

    def test_get_similar_topo(self):
        """Test finding objects with similar topology."""
        result = mtk.get_similar_topo("cyl")
        self.assertEqual(result, [])


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    import importlib

    importlib.reload(mtk.edit_utils)
    mtk.clear_scroll_field_reporter()

    # Create a Test Suite
    suite = unittest.TestSuite()

    # Add the test case class to the suite
    suite.addTest(unittest.makeSuite(EditUtilsTest))

    # Run the suite
    runner = unittest.TextTestRunner(verbosity=2)
    runner.run(suite)


# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------



# !/usr/bin/python
# coding=utf-8
import os
from PySide2 import QtWidgets

try:
    import pymel.core as pm
except ImportError as error:
    print(__file__, error)
import pythontk as ptk

# from this package:
from mayatk.core_utils import CoreUtils
from mayatk.node_utils import NodeUtils
from mayatk.mat_utils import hdr_manager


class StingrayArnoldShader:
    """The StingrayArnoldShader class encapsulates the process of creating a shader network in Maya, specifically aimed
    at bridging Stingray PBS and Arnold aiStandardSurface shaders. It provides a method to create a network of
    interconnected shading nodes based on provided textures, and handles the differences between OpenGL and DirectX
    normal maps by generating the necessary conversions.

    The main method, create_network, takes a list of texture file paths and optional parameters to control the
    naming and behavior of the shader network creation. It ensures necessary plugins are loaded, creates and connects
    shading nodes, sets up the Stingray PBS and Arnold shaders based on the texture types provided, and provides
    feedback through a callback function.

    Error handling is provided to manage missing textures, plugin loading issues, and other potential exceptions that
    might occur during the shader network creation. The method returns early with error feedback if critical issues
    are encountered.

    Attributes:
        The class does not define any new attributes but relies on Maya's PyMel and other utility functions to operate
        on Maya objects and settings.

    Methods:
        create_network(self, textures, name="", normalMapType="OpenGL", callback=print)
            Generates a shader network in Maya based on the provided textures, handling normal map conversions,
            plugin loading, and node connections, while providing feedback through a callback function.
    """

    @CoreUtils.undo
    def create_network(
        self,
        textures,
        name="",
        normalMapType="OpenGL",
        callback=print,
    ):
        """ """
        normal_map_created_from_other_type = False
        # Assure normalMapType is formatted as 'Normal_OpenGL' whether given as 'OpenGL' or 'Normal_OpenGL'
        normalMapType = "Normal_" + normalMapType.strip("Normal_")

        if not textures:
            callback(
                '<br><hl style="color:rgb(255, 100, 100);"><b>Error:</b> No textures given to create_network.</hl>'
            )
            return None
        try:
            pm.loadPlugin("mtoa", quiet=True)  # Assure arnold plugin is loaded.
            # Assure stringray plugin is loaded.
            pm.loadPlugin("shaderFXPlugin", quiet=True)

            sr_node = NodeUtils.create_render_node("StingrayPBS", name=name)
            ai_node = NodeUtils.create_render_node(
                "aiStandardSurface", name=name + "_ai" if name else ""
            )

            opacityMap = ptk.filter_images_by_type(textures, "Opacity")
            if opacityMap:
                maya_install_path = CoreUtils.get_maya_info("install_path")

                graph = os.path.join(
                    maya_install_path,
                    "presets",
                    "ShaderFX",
                    "Scenes",
                    "StingrayPBS",
                    "Standard_Transparent.sfx",
                )
                pm.shaderfx(sfxnode="StingrayPBS1", loadGraph=graph)

            openGLMap = ptk.filter_images_by_type(textures, "Normal_OpenGL")
            directXMap = ptk.filter_images_by_type(textures, "Normal_DirectX")
            if directXMap and not openGLMap and normalMapType == "Normal_OpenGL":
                mapPath = ptk.create_gl_from_dx(directXMap[0])
                textures.append(mapPath)
                normal_map_created_from_other_type = True
                callback(f"OpenGL map created using {ptk.truncate(directXMap[0], 20)}.")
            if openGLMap and not directXMap and normalMapType == "Normal_DirectX":
                mapPath = ptk.create_dx_from_gl(openGLMap[0])
                textures.append(mapPath)
                normal_map_created_from_other_type = True
                callback(f"DirectX map created using {ptk.truncate(openGLMap[0], 20)}.")

            srSG_node = NodeUtils.get_connected_nodes(
                sr_node,
                node_type="shadingEngine",
                direction="outgoing",
                first_match=True,
            )

            aiMult_node = pm.shadingNode("aiMultiply", asShader=True)

            bump_node = pm.shadingNode("bump2d", asShader=True)
            bump_node.bumpInterp.set(1)  # Set bump node to 'tangent space normals'

            NodeUtils.connect_multi_attr(  # Set node connections.
                (ai_node.outColor, srSG_node.aiSurfaceShader),
                (aiMult_node.outColor, ai_node.baseColor),
                (bump_node.outNormal, ai_node.normalCamera),
            )

            length = len(textures)
            progress = 0
            for f in textures:
                typ = ptk.get_image_type_from_filename(f)

                progress += 1

                # Filter normal maps for the correct type.
                if typ == "Normal" and (openGLMap or directXMap):
                    continue
                elif typ == "Normal_DirectX" and normalMapType == "Normal_OpenGL":
                    continue
                elif typ == "Normal_OpenGL" and normalMapType == "Normal_DirectX":
                    continue

                callback(
                    f"creating nodes and connections for <b>{typ}</b> map ..",
                    [progress, length],
                )

                if typ == "Base_Color":
                    n1 = NodeUtils.create_render_node(
                        "file",
                        "as2DTexture",
                        tex=f,
                    )
                    pm.connectAttr(n1.outColor, sr_node.TEX_color_map, force=True)
                    sr_node.use_color_map.set(1)

                    n2 = NodeUtils.create_render_node(
                        "file",
                        "as2DTexture",
                        tex=f,
                        texture_node=True,
                        colorSpace="Raw",
                        ignoreColorSpaceFileRules=1,
                    )
                    pm.connectAttr(n2.outColor, aiMult_node.input1, force=True)

                elif typ == "Roughness":
                    n1 = NodeUtils.create_render_node("file", "as2DTexture", tex=f)
                    pm.connectAttr(n1.outColor, sr_node.TEX_roughness_map, force=True)
                    sr_node.use_roughness_map.set(1)

                    n2 = NodeUtils.create_render_node(
                        "file",
                        "as2DTexture",
                        tex=f,
                        texture_node=True,
                        colorSpace="Raw",
                        alphaIsLuminance=1,
                        ignoreColorSpaceFileRules=1,
                    )
                    pm.connectAttr(n2.outAlpha, ai_node.specularRoughness, force=True)
                    # Opacity: same roughness map used in Specular Roughness to provide additional bluriness of refraction.
                    pm.connectAttr(
                        n2.outAlpha, ai_node.transmissionExtraRoughness, force=True
                    )

                elif typ == "Metallic":
                    n1 = NodeUtils.create_render_node("file", "as2DTexture", tex=f)
                    pm.connectAttr(n1.outColor, sr_node.TEX_metallic_map, force=True)
                    sr_node.use_metallic_map.set(1)

                    n2 = NodeUtils.create_render_node(
                        "file",
                        "as2DTexture",
                        tex=f,
                        texture_node=True,
                        colorSpace="Raw",
                        alphaIsLuminance=1,
                        ignoreColorSpaceFileRules=1,
                    )
                    pm.connectAttr(n2.outAlpha, ai_node.metalness, force=True)

                elif typ == "Emissive":
                    n1 = NodeUtils.create_render_node(
                        "file",
                        "as2DTexture",
                        tex=f,
                    )
                    pm.connectAttr(n1.outColor, sr_node.TEX_emissive_map, force=True)
                    sr_node.use_emissive_map.set(1)

                    n2 = NodeUtils.create_render_node(
                        "file",
                        "as2DTexture",
                        tex=f,
                        colorSpace="Raw",
                        texture_node=True,
                        ignoreColorSpaceFileRules=1,
                    )
                    pm.connectAttr(n2.outAlpha, ai_node.emission, force=True)
                    pm.connectAttr(n2.outColor, ai_node.emissionColor, force=True)

                elif "Normal" in typ:
                    n1 = NodeUtils.create_render_node("file", "as2DTexture", tex=f)
                    pm.connectAttr(n1.outColor, sr_node.TEX_normal_map, force=True)
                    sr_node.use_normal_map.set(1)

                    n2 = NodeUtils.create_render_node(
                        "file",
                        "as2DTexture",
                        tex=f,
                        texture_node=True,
                        colorSpace="Raw",
                        alphaIsLuminance=1,
                        ignoreColorSpaceFileRules=1,
                    )
                    pm.connectAttr(n2.outAlpha, bump_node.bumpValue, force=True)

                elif typ == "Ambient_Occlusion":
                    n1 = NodeUtils.create_render_node(
                        "file",
                        "as2DTexture",
                        tex=f,
                    )
                    pm.connectAttr(n1.outColor, sr_node.TEX_ao_map, force=True)
                    sr_node.use_ao_map.set(1)

                    n2 = NodeUtils.create_render_node(
                        "file",
                        "as2DTexture",
                        tex=f,
                        texture_node=True,
                        colorSpace="Raw",
                        ignoreColorSpaceFileRules=1,
                    )
                    pm.connectAttr(n2.outColor, aiMult_node.input2, force=True)

                elif typ == "Opacity":
                    n1 = NodeUtils.create_render_node("file", "as2DTexture", tex=f)
                    pm.connectAttr(n1.outAlpha, sr_node.opacity, force=True)
                    sr_node.use_opacity_map.set(1)

                    n2 = NodeUtils.create_render_node(
                        "file",
                        "as2DTexture",
                        tex=f,
                        texture_node=True,
                        colorSpace="Raw",
                        alphaIsLuminance=1,
                        ignoreColorSpaceFileRules=1,
                    )
                    pm.connectAttr(n2.outAlpha, ai_node.transmission, force=True)
                    pm.connectAttr(n2.outColor, ai_node.opacity, force=True)

                else:
                    if normal_map_created_from_other_type:
                        # Do not show a warning for unconnected normal maps if it resulted from being converted to a different output type.
                        continue
                    callback(
                        f'<br><hl style="color:rgb(255, 100, 100);"><b>Map type: <b>{typ, ptk.truncate(f, 60)}</b> not connected:<br></hl>',
                        [progress, length],
                    )
                    continue

                callback(
                    f'<font style="color: rgb(80,180,100)">{typ}..connected successfully.</font>'
                )
        except Exception as e:
            callback(
                f'<br><hl style="color:rgb(255, 100, 100);"><b>Error:</b>{e}.</hl>'
            )


class StingrayArnoldShaderSlots(StingrayArnoldShader):
    msg_intro = """<u>To setup the material:</u>
        <br>• Use the <b>Get Texture Maps</b> button to load the images you intend to use.
        <br>• Click the <b>Create Network</b> button to create the shader connections. This will bridge Stingray PBS and Arnold aiStandardSurface shaders, create a shading network from provided textures, and manage OpenGL and DirectX normal map conversions.

        <p><b>Note:</b> To correctly render opacity and transmission, the Opaque setting needs to be disabled on the Shape node.
        If Opaque is enabled, opacity will not work at all. Transmission will work, however any shadows cast by
        the object will always be solid and not pick up the Transparent Color or density of the shader.</p>
    """
    msg_completed = '<br><hl style="color:rgb(0, 255, 255);"><b>COMPLETED.</b></hl>'

    def __init__(self, **kwargs):
        super().__init__()

        self.sb = self.switchboard()
        self.ui = self.sb.stingray_arnold_shader
        self.workspace_dir = CoreUtils.get_maya_info("workspace_dir")
        self.source_images_dir = os.path.join(self.workspace_dir, "sourceimages")
        self.image_files = None

        self.ui.txt001.setText(self.msg_intro)
        self.init_header_menu()

    def init_header_menu(self):
        """Configure header menu"""
        self.ui.header.menu.setTitle("OPTIONS")
        self.ui.header.menu.add(
            self.sb.PushButton,
            setText="HDR Manager",
            setObjectName="b002",
        )

        module = hdr_manager
        slot_class = module.HdrManagerSlots

        # Register and configure HDR Manager UI
        self.sb.register("hdr_manager.ui", slot_class, base_dir=module)
        ui = self.sb.hdr_manager
        ui.set_attributes(WA_TranslucentBackground=True)
        ui.set_flags(FramelessWindowHint=True, WindowStaysOnTopHint=True)
        ui.set_style(theme="dark", style_class="translucentBgWithBorder")
        ui.header.configureButtons(hide_button=True)

        # Connect button click to show HDR Manager
        self.ui.header.menu.b002.clicked.connect(lambda: ui.show(pos="cursor"))

    @property
    def mat_name(self) -> str:
        """Get the mat name from the user input text field.

        Returns:
            (str)
        """
        text = self.ui.txt000.text()
        return text

    @property
    def normal_map_type(self) -> str:
        """Get the normal map type from the comboBoxes current text.

        Returns:
            (str)
        """
        text = self.ui.cmb001.currentText()
        return text

    def b000(self):
        """Create network."""
        if self.image_files:
            # pm.mel.HypershadeWindow() #open the hypershade window.

            self.ui.txt001.clear()
            self.callback("Creating network ..<br>")

            self.create_network(
                self.image_files,
                self.mat_name,
                normalMapType=self.normal_map_type,
                callback=self.callback,
            )

            self.callback(self.msg_completed)
            # pm.mel.hyperShadePanelGraphCommand('hyperShadePanel1', 'rearrangeGraph')

    def b001(self):
        """Get texture maps."""
        image_files = self.sb.file_dialog(
            file_types=["*.png", "*.jpg", "*.bmp", "*.tga", "*.tiff", "*.gif"],
            title="Select one or more image files to open.",
            directory=self.source_images_dir,
        )

        if image_files:
            self.image_files = image_files
            self.ui.txt001.clear()

            msg_mat_selection = self.image_files
            for (
                i
            ) in msg_mat_selection:  # format msg_intro using the map_types in imtools.
                self.callback(ptk.truncate(i, 60))

            self.ui.b000.setDisabled(False)
        elif not self.image_files:
            self.ui.b000.setDisabled(True)

    def callback(self, string, progress=None, clear=False):
        """
        Parameters:
            string (str): The text to output to a textEdit widget.
            progress (int/list): The progress amount to register with the progressBar.
                    Can be given as an int or a tuple as: (progress, total_len)
        """
        if clear:
            self.ui.txt003.clear()

        if isinstance(progress, (list, tuple, set)):
            p, length = progress
            progress = (p / length) * 100

        self.ui.txt001.append(string)

        if progress is not None:
            self.ui.progressBar.setValue(progress)
            QtWidgets.QApplication.instance().processEvents()


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    from uitk import Switchboard

    parent = CoreUtils.get_main_window()
    ui_file = os.path.join(os.path.dirname(__file__), "stingray_arnold_shader.ui")
    sb = Switchboard(
        parent, ui_location=ui_file, slot_location=StingrayArnoldShaderSlots
    )

    sb.current_ui.set_attributes(WA_TranslucentBackground=True)
    sb.current_ui.set_flags(FramelessWindowHint=True, WindowStaysOnTopHint=True)
    sb.current_ui.set_style(theme="dark", style_class="translucentBgWithBorder")
    sb.current_ui.header.configureButtons(
        menu_button=True, minimize_button=True, hide_button=True
    )
    sb.current_ui.show(pos="screen", app_exec=True)

# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------
